# Copyright 2004-2007 by Vinay Sajip. All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of Vinay Sajip
# not be used in advertising or publicity pertaining to distribution
# of the software without specific, written prior permission.
# VINAY SAJIP DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# VINAY SAJIP BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
# ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""
This is a configuration module for Python.

This module should work under Python versions >= 2.2, and cannot be used with
earlier versions since it uses new-style classes.

Development and testing has only been carried out (so far) on Python 2.3.4 and
Python 2.4.2. See the test module (test_config.py) included in the
U{distribution<http://www.red-dove.com/python_config.html|_blank>} (follow the
download link).

A simple example - with the example configuration file::

    messages:
    [
      {
        stream : `sys.stderr`
        message: 'Welcome'
        name: 'Harry'
      }
      {
        stream : `sys.stdout`
        message: 'Welkom'
        name: 'Ruud'
      }
      {
        stream : $messages[0].stream
        message: 'Bienvenue'
        name: Yves
      }
    ]

a program to read the configuration would be::

    from config import Config

    f = file('simple.cfg')
    cfg = Config(f)
    for m in cfg.messages:
        s = '%s, %s' % (m.message, m.name)
        try:
            print >> m.stream, s
        except IOError, e:
            print e

which, when run, would yield the console output::

    Welcome, Harry
    Welkom, Ruud
    Bienvenue, Yves

See U{this tutorial<http://www.red-dove.com/python_config.html|_blank>} for more
information.

@version: 0.3.7

@author: Vinay Sajip

@copyright: Copyright (C) 2004-2007 Vinay Sajip. All Rights Reserved.


@var streamOpener: The default stream opener. This is a factory function which
takes a string (e.g. filename) and returns a stream suitable for reading. If
unable to open the stream, an IOError exception should be thrown.

The default value of this variable is L{defaultStreamOpener}. For an example
of how it's used, see test_config.py (search for streamOpener).
"""

__author__  = "Vinay Sajip <vinay_sajip@red-dove.com>"
__status__  = "alpha"
__version__ = "0.3.7"
__date__    = "05 October 2007"

from types import StringType, UnicodeType

import codecs
import os
import sys

WORD = 'a'
NUMBER = '9'
STRING = '"'
EOF = ''
LCURLY = '{'
RCURLY = '}'
LBRACK = '['
LBRACK2 = 'a['
RBRACK = ']'
LPAREN = '('
LPAREN2 = '(('
RPAREN = ')'
DOT = '.'
COMMA = ','
COLON = ':'
AT = '@'
PLUS = '+'
MINUS = '-'
STAR = '*'
SLASH = '/'
MOD = '%'
BACKTICK = '`'
DOLLAR = '$'
TRUE = 'True'
FALSE = 'False'
NONE = 'None'

WORDCHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_"

if sys.platform == 'win32':
    NEWLINE = '\r\n'
elif os.name == 'mac':
    NEWLINE = '\r'
else:
    NEWLINE = '\n'

try:
    import encodings.utf_32
    encodings.utf_32
    has_utf32 = True
except:
    has_utf32 = False

class ConfigInputStream(object):
    """
    An input stream which can read either ANSI files with default encoding
    or Unicode files with BOMs.

    Handles UTF-8, UTF-16LE, UTF-16BE. Could handle UTF-32 if Python had
    built-in support.
    """
    def __init__(self, stream):
        """
        Initialize an instance.

        @param stream: The underlying stream to be read. Should be seekable.
        @type stream: A stream (file-like object).
        """
        encoding = None
        signature = stream.read(4)
        used = -1
        if has_utf32:
            if signature == codecs.BOM_UTF32_LE:
                encoding = 'utf-32le'
            elif signature == codecs.BOM_UTF32_BE:
                encoding = 'utf-32be'
        if encoding is None:
            if signature[:3] == codecs.BOM_UTF8:
                used = 3
                encoding = 'utf-8'
            elif signature[:2] == codecs.BOM_UTF16_LE:
                used = 2
                encoding = 'utf-16le'
            elif signature[:2] == codecs.BOM_UTF16_BE:
                used = 2
                encoding = 'utf-16be'
            else:
                used = 0
        if used >= 0:
            stream.seek(used)
        if encoding:
            reader = codecs.getreader(encoding)
            stream = reader(stream)
        self.stream = stream
        self.encoding = encoding

    def read(self, size):
        if (size == 0) or (self.encoding is None):
            rv = self.stream.read(size)
        else:
            rv = u''
            while size > 0:
                rv += self.stream.read(1)
                size -= 1
        return rv

    def close(self):
        self.stream.close()

    def readline(self):
        if self.encoding is None:
            line = ''
        else:
            line = u''
        while True:
            c = self.stream.read(1)
            if c:
                line += c
            if c == '\n':
                break
        return line

class ConfigOutputStream(object):
    """
    An output stream which can write either ANSI files with default encoding
    or Unicode files with BOMs.

    Handles UTF-8, UTF-16LE, UTF-16BE. Could handle UTF-32 if Python had
    built-in support.
    """

    def __init__(self, stream, encoding=None):
        """
        Initialize an instance.

        @param stream: The underlying stream to be written.
        @type stream: A stream (file-like object).
        @param encoding: The desired encoding.
        @type encoding: str
        """
        if encoding is not None:
            encoding = str(encoding).lower()
        self.encoding = encoding
        if encoding == "utf-8":
            stream.write(codecs.BOM_UTF8)
        elif encoding == "utf-16be":
            stream.write(codecs.BOM_UTF16_BE)
        elif encoding == "utf-16le":
            stream.write(codecs.BOM_UTF16_LE)
        elif encoding == "utf-32be":
            stream.write(codecs.BOM_UTF32_BE)
        elif encoding == "utf-32le":
            stream.write(codecs.BOM_UTF32_LE)

        if encoding is not None:
            writer = codecs.getwriter(encoding)
            stream = writer(stream)
        self.stream = stream

    def write(self, data):
        self.stream.write(data)

    def flush(self):
        self.stream.flush()

    def close(self):
        self.stream.close()

def defaultStreamOpener(name):
    """
    This function returns a read-only stream, given its name. The name passed
    in should correspond to an existing stream, otherwise an exception will be
    raised.

    This is the default value of L{streamOpener}; assign your own callable to
    streamOpener to return streams based on names. For example, you could use
    urllib2.urlopen().

    @param name: The name of a stream, most commonly a file name.
    @type name: str
    @return: A stream with the specified name.
    @rtype: A read-only stream (file-like object)
    """
    return ConfigInputStream(file(name, 'rb'))

streamOpener = None

class ConfigError(Exception):
    """
    This is the base class of exceptions raised by this module.
    """
    pass

class ConfigFormatError(ConfigError):
    """
    This is the base class of exceptions raised due to syntax errors in
    configurations.
    """
    pass

class ConfigResolutionError(ConfigError):
    """
    This is the base class of exceptions raised due to semantic errors in
    configurations.
    """
    pass

def isWord(s):
    """
    See if a passed-in value is an identifier. If the value passed in is not a
    string, False is returned. An identifier consists of alphanumerics or
    underscore characters.

    Examples::

        isWord('a word') ->False
        isWord('award') -> True
        isWord(9) -> False
        isWord('a_b_c_') ->True

    @note: isWord('9abc') will return True - not exactly correct, but adequate
    for the way it's used here.

    @param s: The name to be tested
    @type s: any
    @return: True if a word, else False
    @rtype: bool
    """
    if type(s) != type(''):
        return False
    s = s.replace('_', '')
    return s.isalnum()

def makePath(prefix, suffix):
    """
    Make a path from a prefix and suffix.

    Examples::

        makePath('', 'suffix') -> 'suffix'
        makePath('prefix', 'suffix') -> 'prefix.suffix'
        makePath('prefix', '[1]') -> 'prefix[1]'

    @param prefix:  The prefix to use. If it evaluates as false, the suffix
                    is returned.
    @type prefix:   str
    @param suffix:  The suffix to use. It is either an identifier or an
                    index in brackets.
    @type suffix:   str
    @return:        The path concatenation of prefix and suffix, with a
                    dot if the suffix is not a bracketed index.
    @rtype:         str

    """
    if not prefix:
        rv = suffix
    elif suffix[0] == '[':
        rv = prefix + suffix
    else:
        rv = prefix + '.' + suffix
    return rv


class Container(object):
    """
    This internal class is the base class for mappings and sequences.

    @ivar path: A string which describes how to get
    to this instance from the root of the hierarchy.

    Example::

        a.list.of[1].or['more'].elements
    """
    def __init__(self, parent):
        """
        Initialize an instance.

        @param parent: The parent of this instance in the hierarchy.
        @type parent: A L{Container} instance.
        """
        object.__setattr__(self, 'parent', parent)

    def setPath(self, path):
        """
        Set the path for this instance.
        @param path: The path - a string which describes how to get
        to this instance from the root of the hierarchy.
        @type path: str
        """
        object.__setattr__(self, 'path', path)

    def evaluate(self, item):
        """
        Evaluate items which are instances of L{Reference} or L{Expression}.

        L{Reference} instances are evaluated using L{Reference.resolve},
        and L{Expression} instances are evaluated using
        L{Expression.evaluate}.

        @param item: The item to be evaluated.
        @type item: any
        @return: If the item is an instance of L{Reference} or L{Expression},
        the evaluated value is returned, otherwise the item is returned
        unchanged.
        """
        if isinstance(item, Reference):
            item = item.resolve(self)
        elif isinstance(item, Expression):
            item = item.evaluate(self)
        return item

    def writeToStream(self, stream, indent, container):
        """
        Write this instance to a stream at the specified indentation level.

        Should be redefined in subclasses.

        @param stream: The stream to write to
        @type stream: A writable stream (file-like object)
        @param indent: The indentation level
        @type indent: int
        @param container: The container of this instance
        @type container: L{Container}
        @raise NotImplementedError: If a subclass does not override this
        """
        raise NotImplementedError

    def writeValue(self, value, stream, indent):
        if isinstance(self, Mapping):
            indstr = ' '
        else:
            indstr = indent * '  '
        if isinstance(value, Reference) or isinstance(value, Expression):
            stream.write('%s%r%s' % (indstr, value, NEWLINE))
        else:
            if (type(value) is StringType): # and not isWord(value):
                value = repr(value)
            stream.write('%s%s%s' % (indstr, value, NEWLINE))

class Mapping(Container):
    """
    This internal class implements key-value mappings in configurations.
    """

    def __init__(self, parent=None):
        """
        Initialize an instance.

        @param parent: The parent of this instance in the hierarchy.
        @type parent: A L{Container} instance.
        """
        Container.__init__(self, parent)
        object.__setattr__(self, 'path', '')
        object.__setattr__(self, 'data', {})
        object.__setattr__(self, 'order', [])   # to preserve ordering
        object.__setattr__(self, 'comments', {})

    def __delitem__(self, key):
        """
        Remove an item
        """
        data = object.__getattribute__(self, 'data')
        if key not in data:
            raise AttributeError(key)
        order = object.__getattribute__(self, 'order')
        comments = object.__getattribute__(self, 'comments')
        del data[key]
        order.remove(key)
        del comments[key]

    def __getitem__(self, key):
        data = object.__getattribute__(self, 'data')
        if key not in data:
            raise AttributeError(key)
        rv = data[key]
        return self.evaluate(rv)

    __getattr__ = __getitem__

    def __getattribute__(self, name):
        if name == "__dict__":
            return {}
        if name in ["__methods__", "__members__"]:
            return []
        #if name == "__class__":
        #    return ''
        data = object.__getattribute__(self, "data")
        useData = data.has_key(name)
        if useData:
            rv = getattr(data, name)
        else:
            rv = object.__getattribute__(self, name)
            if rv is None:
                raise AttributeError(name)
        return rv

    def iteritems(self):
        for key in self.keys():
            yield(key, self[key])
        raise StopIteration

    def __contains__(self, item):
        order = object.__getattribute__(self, 'order')
        return item in order

    def addMapping(self, key, value, comment, setting=False):
        """
        Add a key-value mapping with a comment.

        @param key: The key for the mapping.
        @type key: str
        @param value: The value for the mapping.
        @type value: any
        @param comment: The comment for the key (can be None).
        @type comment: str
        @param setting: If True, ignore clashes. This is set
        to true when called from L{__setattr__}.
        @raise ConfigFormatError: If an existing key is seen
        again and setting is False.
        """
        data = object.__getattribute__(self, 'data')
        order = object.__getattribute__(self, 'order')
        comments = object.__getattribute__(self, 'comments')

        data[key] = value
        if key not in order:
            order.append(key)
        elif not setting:
            raise ConfigFormatError("repeated key: %s" % key)
        comments[key] = comment

    def __setattr__(self, name, value):
        self.addMapping(name, value, None, True)

    __setitem__ = __setattr__

    def keys(self):
        """
        Return the keys in a similar way to a dictionary.
        """
        return object.__getattribute__(self, 'order')

    def get(self, key, default=None):
        """
        Allows a dictionary-style get operation.
        """
        if key in self:
            return self[key]
        return default

    def __str__(self):
        return str(object.__getattribute__(self, 'data'))

    def __repr__(self):
        return repr(object.__getattribute__(self, 'data'))

    def __len__(self):
        return len(object.__getattribute__(self, 'order'))

    def __iter__(self):
        return self.iterkeys()

    def iterkeys(self):
        order = object.__getattribute__(self, 'order')
        return order.__iter__()

    def writeToStream(self, stream, indent, container):
        """
        Write this instance to a stream at the specified indentation level.

        Should be redefined in subclasses.

        @param stream: The stream to write to
        @type stream: A writable stream (file-like object)
        @param indent: The indentation level
        @type indent: int
        @param container: The container of this instance
        @type container: L{Container}
        """
        indstr = indent * '  '
        if len(self) == 0:
            stream.write(' { }%s' % NEWLINE)
        else:
            if isinstance(container, Mapping):
                stream.write(NEWLINE)
            stream.write('%s{%s' % (indstr, NEWLINE))
            self.save(stream, indent + 1)
            stream.write('%s}%s' % (indstr, NEWLINE))

    def save(self, stream, indent=0):
        """
        Save this configuration to the specified stream.
        @param stream: A stream to which the configuration is written.
        @type stream: A write-only stream (file-like object).
        @param indent: The indentation level for the output.
        @type indent: int
        """
        indstr = indent * '  '
        order = object.__getattribute__(self, 'order')
        data = object.__getattribute__(self, 'data')
        maxlen = 0 # max(map(lambda x: len(x), order))
        for key in order:
            comment = self.comments[key]
            if isWord(key):
                skey = key
            else:
                skey = repr(key)
            if comment:
                stream.write('%s#%s' % (indstr, comment))
            stream.write('%s%-*s :' % (indstr, maxlen, skey))
            value = data[key]
            if isinstance(value, Container):
                value.writeToStream(stream, indent, self)
            else:
                self.writeValue(value, stream, indent)

class Config(Mapping):
    """
    This class represents a configuration, and is the only one which clients
    need to interface to, under normal circumstances.
    """

    class Namespace(object):
        """
        This internal class is used for implementing default namespaces.

        An instance acts as a namespace.
        """
        def __init__(self):
            self.sys = sys
            self.os = os

    def __init__(self, streamOrFile=None, parent=None):
        """
        Initializes an instance.

        @param streamOrFile: If specified, causes this instance to be loaded
        from the stream (by calling L{load}). If a string is provided, it is
        passed to L{streamOpener} to open a stream. Otherwise, the passed
        value is assumed to be a stream and used as is.
        @type streamOrFile: A readable stream (file-like object) or a name.
        @param parent: If specified, this becomes the parent of this instance
        in the configuration hierarchy.
        @type parent: a L{Container} instance.
        """
        Mapping.__init__(self, parent)
        object.__setattr__(self, 'reader', ConfigReader(self))
        object.__setattr__(self, 'namespaces', [Config.Namespace()])
        if streamOrFile is not None:
            if isinstance(streamOrFile, StringType) or isinstance(streamOrFile, UnicodeType):
                global streamOpener
                if streamOpener is None:
                    streamOpener = defaultStreamOpener
                streamOrFile = streamOpener(streamOrFile)
            load = object.__getattribute__(self, "load")
            load(streamOrFile)

    def load(self, stream):
        """
        Load the configuration from the specified stream. Multiple streams can
        be used to populate the same instance, as long as there are no
        clashing keys. The stream is closed.
        @param stream: A stream from which the configuration is read.
        @type stream: A read-only stream (file-like object).
        @raise ConfigError: if keys in the loaded configuration clash with
        existing keys.
        @raise ConfigFormatError: if there is a syntax error in the stream.
        """
        reader = object.__getattribute__(self, 'reader')
        #object.__setattr__(self, 'root', reader.load(stream))
        reader.load(stream)
        stream.close()

    def addNamespace(self, ns, name=None):
        """
        Add a namespace to this configuration which can be used to evaluate
        (resolve) dotted-identifier expressions.
        @param ns: The namespace to be added.
        @type ns: A module or other namespace suitable for passing as an
        argument to vars().
        @param name: A name for the namespace, which, if specified, provides
        an additional level of indirection.
        @type name: str
        """
        namespaces = object.__getattribute__(self, 'namespaces')
        if name is None:
            namespaces.append(ns)
        else:
            setattr(namespaces[0], name, ns)

    def removeNamespace(self, ns, name=None):
        """
        Remove a namespace added with L{addNamespace}.
        @param ns: The namespace to be removed.
        @param name: The name which was specified when L{addNamespace} was
        called.
        @type name: str
        """
        namespaces = object.__getattribute__(self, 'namespaces')
        if name is None:
            namespaces.remove(ns)
        else:
            delattr(namespaces[0], name)

    def save(self, stream, indent=0):
        """
        Save this configuration to the specified stream. The stream is
        closed if this is the top-level configuration in the hierarchy.
        L{Mapping.save} is called to do all the work.
        @param stream: A stream to which the configuration is written.
        @type stream: A write-only stream (file-like object).
        @param indent: The indentation level for the output.
        @type indent: int
        """
        Mapping.save(self, stream, indent)
        if indent == 0:
            stream.close()

    def getByPath(self, path):
        """
        Obtain a value in the configuration via its path.
        @param path: The path of the required value
        @type path: str
        @return the value at the specified path.
        @rtype: any
        @raise ConfigError: If the path is invalid
        """
        s = 'self.' + path
        try:
            return eval(s)
        except Exception, e:
            raise ConfigError(str(e))

class Sequence(Container):
    """
    This internal class implements a value which is a sequence of other values.
    """
    class SeqIter(object):
        """
        This internal class implements an iterator for a L{Sequence} instance.
        """
        def __init__(self, seq):
            self.seq = seq
            self.limit = len(object.__getattribute__(seq, 'data'))
            self.index = 0

        def __iter__(self):
            return self

        def next(self):
            if self.index >= self.limit:
                raise StopIteration
            rv = self.seq[self.index]
            self.index += 1
            return rv

    def __init__(self, parent=None):
        """
        Initialize an instance.

        @param parent: The parent of this instance in the hierarchy.
        @type parent: A L{Container} instance.
        """
        Container.__init__(self, parent)
        object.__setattr__(self, 'data', [])
        object.__setattr__(self, 'comments', [])

    def append(self, item, comment):
        """
        Add an item to the sequence.

        @param item: The item to add.
        @type item: any
        @param comment: A comment for the item.
        @type comment: str
        """
        data = object.__getattribute__(self, 'data')
        comments = object.__getattribute__(self, 'comments')
        data.append(item)
        comments.append(comment)

    def __getitem__(self, index):
        data = object.__getattribute__(self, 'data')
        try:
            rv = data[index]
        except (IndexError, KeyError, TypeError):
            raise ConfigResolutionError('%r is not a valid index for %r' % (index, object.__getattribute__(self, 'path')))
        if not isinstance(rv, list):
            rv = self.evaluate(rv)
        else:
            # deal with a slice
            result = []
            for a in rv:
                result.append(self.evaluate(a))
            rv = result
        return rv

    def __iter__(self):
        return Sequence.SeqIter(self)

    def __repr__(self):
        return repr(object.__getattribute__(self, 'data'))

    def __str__(self):
        return str(self[:]) # using the slice evaluates the contents

    def __len__(self):
        return len(object.__getattribute__(self, 'data'))

    def writeToStream(self, stream, indent, container):
        """
        Write this instance to a stream at the specified indentation level.

        Should be redefined in subclasses.

        @param stream: The stream to write to
        @type stream: A writable stream (file-like object)
        @param indent: The indentation level
        @type indent: int
        @param container: The container of this instance
        @type container: L{Container}
        """
        indstr = indent * '  '
        if len(self) == 0:
            stream.write(' [ ]%s' % NEWLINE)
        else:
            if isinstance(container, Mapping):
                stream.write(NEWLINE)
            stream.write('%s[%s' % (indstr, NEWLINE))
            self.save(stream, indent + 1)
            stream.write('%s]%s' % (indstr, NEWLINE))

    def save(self, stream, indent):
        """
        Save this instance to the specified stream.
        @param stream: A stream to which the configuration is written.
        @type stream: A write-only stream (file-like object).
        @param indent: The indentation level for the output, > 0
        @type indent: int
        """
        if indent == 0:
            raise ConfigError("sequence cannot be saved as a top-level item")
        data = object.__getattribute__(self, 'data')
        comments = object.__getattribute__(self, 'comments')
        indstr = indent * '  '
        for i in xrange(0, len(data)):
            value = data[i]
            comment = comments[i]
            if comment:
                stream.write('%s#%s' % (indstr, comment))
            if isinstance(value, Container):
                value.writeToStream(stream, indent, self)
            else:
                self.writeValue(value, stream, indent)

class Reference(object):
    """
    This internal class implements a value which is a reference to another value.
    """
    def __init__(self, config, type, ident):
        """
        Initialize an instance.

        @param config: The configuration which contains this reference.
        @type config: A L{Config} instance.
        @param type: The type of reference.
        @type type: BACKTICK or DOLLAR
        @param ident: The identifier which starts the reference.
        @type ident: str
        """
        self.config = config
        self.type = type
        self.elements = [ident]

    def addElement(self, type, ident):
        """
        Add an element to the reference.

        @param type: The type of reference.
        @type type: BACKTICK or DOLLAR
        @param ident: The identifier which continues the reference.
        @type ident: str
        """
        self.elements.append((type, ident))

    def findConfig(self, container):
        """
        Find the closest enclosing configuration to the specified container.

        @param container: The container to start from.
        @type container: L{Container}
        @return: The closest enclosing configuration, or None.
        @rtype: L{Config}
        """
        while (container is not None) and not isinstance(container, Config):
            container = object.__getattribute__(container, 'parent')
        return container

    def resolve(self, container):
        """
        Resolve this instance in the context of a container.

        @param container: The container to resolve from.
        @type container: L{Container}
        @return: The resolved value.
        @rtype: any
        @raise ConfigResolutionError: If resolution fails.
        """
        rv = None
        path = object.__getattribute__(container, 'path')
        current = self.findConfig(container)
        while current is not None:
            if self.type == BACKTICK:
                namespaces = object.__getattribute__(current, 'namespaces')
                found = False
                for ns in namespaces:
                    try:
                        rv = eval(str(self)[1:-1], vars(ns))
                        found = True
                        break
                    except:
                        pass
                if found:
                    break
            else:
                key = self.elements[0]
                try:
                    rv = current[key]
                    for item in self.elements[1:]:
                        key = item[1]
                        rv = rv[key]
                    break
                except:
                    rv = None
                    pass
            current = self.findConfig(object.__getattribute__(current, 'parent'))
        if current is None:
            raise ConfigResolutionError("unable to evaluate %r in the configuration %s" % (self, path))
        return rv

    def __str__(self):
        s = self.elements[0]
        for tt, tv in self.elements[1:]:
            if tt == DOT:
                s += '.%s' % tv
            else:
                s += '[%r]' % tv
        if self.type == BACKTICK:
            return BACKTICK + s + BACKTICK
        else:
            return DOLLAR + s

    def __repr__(self):
        return self.__str__()

class Expression(object):
    """
    This internal class implements a value which is obtained by evaluating an expression.
    """
    def __init__(self, op, lhs, rhs):
        """
        Initialize an instance.

        @param op: the operation expressed in the expression.
        @type op: PLUS, MINUS, STAR, SLASH, MOD
        @param lhs: the left-hand-side operand of the expression.
        @type lhs: any Expression or primary value.
        @param rhs: the right-hand-side operand of the expression.
        @type rhs: any Expression or primary value.
        """
        self.op = op
        self.lhs = lhs
        self.rhs = rhs

    def __str__(self):
        return '%r %s %r' % (self.lhs, self.op, self.rhs)

    def __repr__(self):
        return self.__str__()

    def evaluate(self, container):
        """
        Evaluate this instance in the context of a container.

        @param container: The container to evaluate in from.
        @type container: L{Container}
        @return: The evaluated value.
        @rtype: any
        @raise ConfigResolutionError: If evaluation fails.
        @raise ZeroDivideError: If division by zero occurs.
        @raise TypeError: If the operation is invalid, e.g.
        subtracting one string from another.
        """
        lhs = self.lhs
        if isinstance(lhs, Reference):
            lhs = lhs.resolve(container)
        elif isinstance(lhs, Expression):
            lhs = lhs.evaluate(container)
        rhs = self.rhs
        if isinstance(rhs, Reference):
            rhs = rhs.resolve(container)
        elif isinstance(rhs, Expression):
            rhs = rhs.evaluate(container)
        op = self.op
        if op == PLUS:
            rv = lhs + rhs
        elif op == MINUS:
            rv = lhs - rhs
        elif op == STAR:
            rv = lhs * rhs
        elif op == SLASH:
            rv = lhs / rhs
        else:
            rv = lhs % rhs
        return rv

class ConfigReader(object):
    """
    This internal class implements a parser for configurations.
    """

    def __init__(self, config):
        self.filename = None
        self.config = config
        self.lineno = 0
        self.colno = 0
        self.lastc = None
        self.last_token = None
        self.commentchars = '#'
        self.whitespace = ' \t\r\n'
        self.quotes = '\'"'
        self.punct = ':-+*/%,.{}[]()@`$'
        self.digits = '0123456789'
        self.wordchars = '%s' % WORDCHARS # make a copy
        self.identchars = self.wordchars + self.digits
        self.pbchars = []
        self.pbtokens = []
        self.comment = None

    def location(self):
        """
        Return the current location (filename, line, column) in the stream
        as a string.

        Used when printing error messages,

        @return: A string representing a location in the stream being read.
        @rtype: str
        """
        return "%s(%d,%d)" % (self.filename, self.lineno, self.colno)

    def getChar(self):
        """
        Get the next char from the stream. Update line and column numbers
        appropriately.

        @return: The next character from the stream.
        @rtype: str
        """
        if self.pbchars:
            c = self.pbchars.pop()
        else:
            c = self.stream.read(1)
            self.colno += 1
            if c == '\n':
                self.lineno += 1
                self.colno = 1
        return c

    def __repr__(self):
        return "<ConfigReader at 0x%08x>" % id(self)

    __str__ = __repr__

    def getToken(self):
        """
        Get a token from the stream. String values are returned in a form
        where you need to eval() the returned value to get the actual
        string. The return value is (token_type, token_value).

        Multiline string tokenizing is thanks to David Janes (BlogMatrix)

        @return: The next token.
        @rtype: A token tuple.
        """
        if self.pbtokens:
            return self.pbtokens.pop()
        stream = self.stream
        self.comment = None
        token = ''
        tt = EOF
        while True:
            c = self.getChar()
            if not c:
                break
            elif c == '#':
                self.comment = stream.readline()
                self.lineno += 1
                continue
            if c in self.quotes:
                token = c
                quote = c
                tt = STRING
                escaped = False
                multiline = False
                c1 = self.getChar()
                if c1 == quote:
                    c2 = self.getChar()
                    if c2 == quote:
                        multiline = True
                        token += quote
                        token += quote
                    else:
                        self.pbchars.append(c2)
                        self.pbchars.append(c1)
                else:
                    self.pbchars.append(c1)
                while True:
                    c = self.getChar()
                    if not c:
                        break
                    token += c
                    if (c == quote) and not escaped:
                        if not multiline or (len(token) >= 6 and token.endswith(token[:3]) and token[-4] != '\\'):
                            break
                    if c == '\\':
                        escaped = not escaped
                    else:
                        escaped = False
                if not c:
                    raise ConfigFormatError('%s: Unterminated quoted string: %r, %r' % (self.location(), token, c))
                break
            if c in self.whitespace:
                self.lastc = c
                continue
            elif c in self.punct:
                token = c
                tt = c
                if (self.lastc == ']') or (self.lastc in self.identchars):
                    if c == '[':
                        tt = LBRACK2
                    elif c == '(':
                        tt = LPAREN2
                break
            elif c in self.digits:
                token = c
                tt = NUMBER
                while True:
                    c = self.getChar()
                    if not c:
                        break
                    if c in self.digits:
                        token += c
                    elif (c == '.') and token.find('.') < 0:
                        token += c
                    else:
                        if c and (c not in self.whitespace):
                            self.pbchars.append(c)
                        break
                break
            elif c in self.wordchars:
                token = c
                tt = WORD
                c = self.getChar()
                while c and (c in self.identchars):
                    token += c
                    c = self.getChar()
                if c: # and c not in self.whitespace:
                    self.pbchars.append(c)
                if token == "True":
                    tt = TRUE
                elif token == "False":
                    tt = FALSE
                elif token == "None":
                    tt = NONE
                break
            else:
                raise ConfigFormatError('%s: Unexpected character: %r' % (self.location(), c))
        if token:
            self.lastc = token[-1]
        else:
            self.lastc = None
        self.last_token = tt
        return (tt, token)

    def load(self, stream, parent=None, suffix=None):
        """
        Load the configuration from the specified stream.

        @param stream: A stream from which to load the configuration.
        @type stream: A stream (file-like object).
        @param parent: The parent of the configuration (to which this reader
        belongs) in the hierarchy. Specified when the configuration is
        included in another one.
        @type parent: A L{Container} instance.
        @param suffix: The suffix of this configuration in the parent
        configuration. Should be specified whenever the parent is not None.
        @raise ConfigError: If parent is specified but suffix is not.
        @raise ConfigFormatError: If there are syntax errors in the stream.
        """
        if parent is not None:
            if suffix is None:
                raise ConfigError("internal error: load called with parent but no suffix")
            self.config.setPath(makePath(object.__getattribute__(parent, 'path'), suffix))
        self.setStream(stream)
        self.token = self.getToken()
        self.parseMappingBody(self.config)
        if self.token[0] != EOF:
            raise ConfigFormatError('%s: expecting EOF, found %r' % (self.location(), self.token[1]))

    def setStream(self, stream):
        """
        Set the stream to the specified value, and prepare to read from it.

        @param stream: A stream from which to load the configuration.
        @type stream: A stream (file-like object).
        """
        self.stream = stream
        if hasattr(stream, 'name'):
            filename = stream.name
        else:
            filename = '?'
        self.filename = filename
        self.lineno = 1
        self.colno = 1

    def match(self, t):
        """
        Ensure that the current token type matches the specified value, and
        advance to the next token.

        @param t: The token type to match.
        @type t: A valid token type.
        @return: The token which was last read from the stream before this
        function is called.
        @rtype: a token tuple - see L{getToken}.
        @raise ConfigFormatError: If the token does not match what's expected.
        """
        if self.token[0] != t:
            raise ConfigFormatError("%s: expecting %s, found %r" % (self.location(), t, self.token[1]))
        rv = self.token
        self.token = self.getToken()
        return rv

    def parseMappingBody(self, parent):
        """
        Parse the internals of a mapping, and add entries to the provided
        L{Mapping}.

        @param parent: The mapping to add entries to.
        @type parent: A L{Mapping} instance.
        """
        while self.token[0] in [WORD, STRING]:
            self.parseKeyValuePair(parent)

    def parseKeyValuePair(self, parent):
        """
        Parse a key-value pair, and add it to the provided L{Mapping}.

        @param parent: The mapping to add entries to.
        @type parent: A L{Mapping} instance.
        @raise ConfigFormatError: if a syntax error is found.
        """
        comment = self.comment
        tt, tv = self.token
        if tt == WORD:
            key = tv
            suffix = tv
        elif tt == STRING:
            key = eval(tv)
            suffix = '[%s]' % tv
        else:
            msg = "%s: expecting word or string, found %r"
            raise ConfigFormatError(msg % (self.location(), tv))
        self.token = self.getToken()
        # for now, we allow key on its own as a short form of key : True
        if self.token[0] == COLON:
            self.token = self.getToken()
            value = self.parseValue(parent, suffix)
        else:
            value = True
        try:
            parent.addMapping(key, value, comment)
        except Exception, e:
            raise ConfigFormatError("%s: %s, %r" % (self.location(), e,
                                    self.token[1]))
        tt = self.token[0]
        if tt not in [EOF, WORD, STRING, RCURLY, COMMA]:
            msg = "%s: expecting one of EOF, WORD, STRING,\
RCURLY, COMMA, found %r"
            raise ConfigFormatError(msg  % (self.location(), self.token[1]))
        if tt == COMMA:
            self.token = self.getToken()

    def parseValue(self, parent, suffix):
        """
        Parse a value.

        @param parent: The container to which the value will be added.
        @type parent: A L{Container} instance.
        @param suffix: The suffix for the value.
        @type suffix: str
        @return: The value
        @rtype: any
        @raise ConfigFormatError: if a syntax error is found.
        """
        tt = self.token[0]
        if tt in [STRING, WORD, NUMBER, LPAREN, DOLLAR,
                  TRUE, FALSE, NONE, BACKTICK, MINUS]:
            rv = self.parseScalar()
        elif tt == LBRACK:
            rv = self.parseSequence(parent, suffix)
        elif tt in [LCURLY, AT]:
            rv = self.parseMapping(parent, suffix)
        else:
            raise ConfigFormatError("%s: unexpected input: %r" %
               (self.location(), self.token[1]))
        return rv

    def parseSequence(self, parent, suffix):
        """
        Parse a sequence.

        @param parent: The container to which the sequence will be added.
        @type parent: A L{Container} instance.
        @param suffix: The suffix for the value.
        @type suffix: str
        @return: a L{Sequence} instance representing the sequence.
        @rtype: L{Sequence}
        @raise ConfigFormatError: if a syntax error is found.
        """
        rv = Sequence(parent)
        rv.setPath(makePath(object.__getattribute__(parent, 'path'), suffix))
        self.match(LBRACK)
        comment = self.comment
        tt = self.token[0]
        while tt in [STRING, WORD, NUMBER, LCURLY, LBRACK, LPAREN, DOLLAR,
                     TRUE, FALSE, NONE, BACKTICK]:
            suffix = '[%d]' % len(rv)
            value = self.parseValue(parent, suffix)
            rv.append(value, comment)
            tt = self.token[0]
            comment = self.comment
            if tt == COMMA:
                self.match(COMMA)
                tt = self.token[0]
                comment = self.comment
                continue
        self.match(RBRACK)
        return rv

    def parseMapping(self, parent, suffix):
        """
        Parse a mapping.

        @param parent: The container to which the mapping will be added.
        @type parent: A L{Container} instance.
        @param suffix: The suffix for the value.
        @type suffix: str
        @return: a L{Mapping} instance representing the mapping.
        @rtype: L{Mapping}
        @raise ConfigFormatError: if a syntax error is found.
        """
        if self.token[0] == LCURLY:
            self.match(LCURLY)
            rv = Mapping(parent)
            rv.setPath(
               makePath(object.__getattribute__(parent, 'path'), suffix))
            self.parseMappingBody(rv)
            self.match(RCURLY)
        else:
            self.match(AT)
            tt, fn = self.match(STRING)
            rv = Config(eval(fn), parent)
        return rv

    def parseScalar(self):
        """
        Parse a scalar - a terminal value such as a string or number, or
        an L{Expression} or L{Reference}.

        @return: the parsed scalar
        @rtype: any scalar
        @raise ConfigFormatError: if a syntax error is found.
        """
        lhs = self.parseTerm()
        tt = self.token[0]
        while tt in [PLUS, MINUS]:
            self.match(tt)
            rhs = self.parseTerm()
            lhs = Expression(tt, lhs, rhs)
            tt = self.token[0]
        return lhs

    def parseTerm(self):
        """
        Parse a term in an additive expression (a + b, a - b)

        @return: the parsed term
        @rtype: any scalar
        @raise ConfigFormatError: if a syntax error is found.
        """
        lhs = self.parseFactor()
        tt = self.token[0]
        while tt in [STAR, SLASH, MOD]:
            self.match(tt)
            rhs = self.parseFactor()
            lhs = Expression(tt, lhs, rhs)
            tt = self.token[0]
        return lhs

    def parseFactor(self):
        """
        Parse a factor in an multiplicative expression (a * b, a / b, a % b)

        @return: the parsed factor
        @rtype: any scalar
        @raise ConfigFormatError: if a syntax error is found.
        """
        tt = self.token[0]
        if tt in [NUMBER, WORD, STRING, TRUE, FALSE, NONE]:
            rv = self.token[1]
            if tt != WORD:
                rv = eval(rv)
            self.match(tt)
        elif tt == LPAREN:
            self.match(LPAREN)
            rv = self.parseScalar()
            self.match(RPAREN)
        elif tt == DOLLAR:
            self.match(DOLLAR)
            rv = self.parseReference(DOLLAR)
        elif tt == BACKTICK:
            self.match(BACKTICK)
            rv = self.parseReference(BACKTICK)
            self.match(BACKTICK)
        elif tt == MINUS:
            self.match(MINUS)
            rv = -self.parseScalar()
        else:
            raise ConfigFormatError("%s: unexpected input: %r" %
               (self.location(), self.token[1]))
        return rv

    def parseReference(self, type):
        """
        Parse a reference.

        @return: the parsed reference
        @rtype: L{Reference}
        @raise ConfigFormatError: if a syntax error is found.
        """
        word = self.match(WORD)
        rv = Reference(self.config, type, word[1])
        while self.token[0] in [DOT, LBRACK2]:
            self.parseSuffix(rv)
        return rv

    def parseSuffix(self, ref):
        """
        Parse a reference suffix.

        @param ref: The reference of which this suffix is a part.
        @type ref: L{Reference}.
        @raise ConfigFormatError: if a syntax error is found.
        """
        tt = self.token[0]
        if tt == DOT:
            self.match(DOT)
            word = self.match(WORD)
            ref.addElement(DOT, word[1])
        else:
            self.match(LBRACK2)
            tt, tv = self.token
            if tt not in [NUMBER, STRING]:
                raise ConfigFormatError("%s: expected number or string, found %r" % (self.location(), tv))
            self.token = self.getToken()
            tv = eval(tv)
            self.match(RBRACK)
            ref.addElement(LBRACK, tv)

def defaultMergeResolve(map1, map2, key):
    """
    A default resolver for merge conflicts. Returns a string
    indicating what action to take to resolve the conflict.

    @param map1: The map being merged into.
    @type map1: L{Mapping}.
    @param map2: The map being used as the merge operand.
    @type map2: L{Mapping}.
    @param key: The key in map2 (which also exists in map1).
    @type key: str
    @return: One of "merge", "append", "mismatch" or "overwrite"
             indicating what action should be taken. This should
             be appropriate to the objects being merged - e.g.
             there is no point returning "merge" if the two objects
             are instances of L{Sequence}.
    @rtype: str
    """
    obj1 = map1[key]
    obj2 = map2[key]
    if isinstance(obj1, Mapping) and isinstance(obj2, Mapping):
        rv = "merge"
    elif isinstance(obj1, Sequence) and isinstance(obj2, Sequence):
        rv = "append"
    else:
        rv = "mismatch"
    return rv

def overwriteMergeResolve(map1, map2, key):
    """
    An overwriting resolver for merge conflicts. Calls L{defaultMergeResolve},
    but where a "mismatch" is detected, returns "overwrite" instead.

    @param map1: The map being merged into.
    @type map1: L{Mapping}.
    @param map2: The map being used as the merge operand.
    @type map2: L{Mapping}.
    @param key: The key in map2 (which also exists in map1).
    @type key: str
    """
    rv = defaultMergeResolve(map1, map2, key)
    if rv == "mismatch":
        rv = "overwrite"
    return rv

class ConfigMerger(object):
    """
    This class is used for merging two configurations. If a key exists in the
    merge operand but not the merge target, then the entry is copied from the
    merge operand to the merge target. If a key exists in both configurations,
    then a resolver (a callable) is called to decide how to handle the
    conflict.
    """

    def __init__(self, resolver=defaultMergeResolve):
        """
        Initialise an instance.

        @param resolver:
        @type resolver: A callable which takes the argument list
        (map1, map2, key) where map1 is the mapping being merged into,
        map2 is the merge operand and key is the clashing key. The callable
        should return a string indicating how the conflict should be resolved.
        For possible return values, see L{defaultMergeResolve}. The default
        value preserves the old behaviour
        """
        self.resolver = resolver

    def merge(self, merged, mergee):
        """
        Merge two configurations. The second configuration is unchanged,
        and the first is changed to reflect the results of the merge.

        @param merged: The configuration to merge into.
        @type merged: L{Config}.
        @param mergee: The configuration to merge.
        @type mergee: L{Config}.
        """
        self.mergeMapping(merged, mergee)

    def mergeMapping(self, map1, map2):
        """
        Merge two mappings recursively. The second mapping is unchanged,
        and the first is changed to reflect the results of the merge.

        @param map1: The mapping to merge into.
        @type map1: L{Mapping}.
        @param map2: The mapping to merge.
        @type map2: L{Mapping}.
        """
        keys = map1.keys()
        for key in map2.keys():
            if key not in keys:
                map1[key] = map2[key]
            else:
                obj1 = map1[key]
                obj2 = map2[key]
                decision = self.resolver(map1, map2, key)
                if decision == "merge":
                    self.mergeMapping(obj1, obj2)
                elif decision == "append":
                    self.mergeSequence(obj1, obj2)
                elif decision == "overwrite":
                    map1[key] = obj2
                elif decision == "mismatch":
                    self.handleMismatch(obj1, obj2)
                else:
                    msg = "unable to merge: don't know how to implement %r"
                    raise ValueError(msg % decision)

    def mergeSequence(self, seq1, seq2):
        """
        Merge two sequences. The second sequence is unchanged,
        and the first is changed to have the elements of the second
        appended to it.

        @param seq1: The sequence to merge into.
        @type seq1: L{Sequence}.
        @param seq2: The sequence to merge.
        @type seq2: L{Sequence}.
        """
        data1 = object.__getattribute__(seq1, 'data')
        data2 = object.__getattribute__(seq2, 'data')
        for obj in data2:
            data1.append(obj)
        comment1 = object.__getattribute__(seq1, 'comments')
        comment2 = object.__getattribute__(seq2, 'comments')
        for obj in comment2:
            comment1.append(obj)

    def handleMismatch(self, obj1, obj2):
        """
        Handle a mismatch between two objects.

        @param obj1: The object to merge into.
        @type obj1: any
        @param obj2: The object to merge.
        @type obj2: any
        """
        raise ConfigError("unable to merge %r with %r" % (obj1, obj2))

class ConfigList(list):
    """
    This class implements an ordered list of configurations and allows you
    to try getting the configuration from each entry in turn, returning
    the first successfully obtained value.
    """

    def getByPath(self, path):
        """
        Obtain a value from the first configuration in the list which defines
        it.

        @param path: The path of the value to retrieve.
        @type path: str
        @return: The value from the earliest configuration in the list which
        defines it.
        @rtype: any
        @raise ConfigError: If no configuration in the list has an entry with
        the specified path.
        """
        found = False
        rv = None
        for entry in self:
            try:
                rv = entry.getByPath(path)
                found = True
                break
            except ConfigError:
                pass
        if not found:
            raise ConfigError("unable to resolve %r" % path)
        return rv



"""Python executable-serving system

ReadyExec is a client-server system designed to alleviate the
problem of high-startup-costing applications which are run
repeatedly (e.g., in procmail), and use stdio files, argv,
and exit codes interact can be used within ReadyExec with very
little work.

Introduction

There are two components to the ReadyExec system.
There is the readyexecd.py server and the readyexec client.


readyexecd.py:

  usage: readyexecd.py socket_path callable_object

readyexecd.py is a long-running daemon process that listens
on a unix domain socket, socket_path.  When it receives
requests on socket_path, it then executes callable_object,
which is the path to a Python callable object, e.g.,
module.function

While executing the callable object, it connects its
stdio file streams to the client, so that any standard
input, output, and error is able to be read in by the
completely unaware module.function.

See examples later in this document.

Note: readyexecd.py does not daemonize itself.


readyexec:

  usage: readyexec [--stop] socket_path program_args

readyexec is a short-lived client process that sends
a request to execute the code that the readyexecd.py listening
on socket_path is serving.  It sends its args, environment,
and stdio file descriptors to the server, and then
reads back an error code.

If --stop is given, it tells the server to shutdown.

See examples later in this document.


How to get your Python script using ReadyExec:

If you have a script that would like to make use of ReadyExec,
merely make a module that can be imported containing a
function which runs your script-specfic code.
Your sys.argv and sys.std* filehandles will be replaced
in the background, so users running readyexec will be
able to communicate with your process in a 'normal' fashion.
Also, if you raise an exception or call sys.exit, readyexec
will exit with the appropriate exit code (else exiting with 0)..

Note that you also have access to readyexec's environment variables.


Examples

Assuming that mymodule.foobar is a Python function:

  readyexecd.py /tmp/foobar mymodule.foobar

Now, we can 'call' mymodule.foobar, passing in 'args'
(note that these are not arguments to the function, but
rather just put into the argument vector sys.argv
that mymodule.foobar can read)

  readyexec /tmp/foobar
  readyexec /tmp/foobar arg1 arg2
  readyexec /tmp/foobar < input > output


Author

   Frank J. Tobin, ftobin@neverending.org

   OpenPGP key fingerprint:
   4F86 3BBB A816 6F0A 340F  6003 56FF D10A 260C 4FA3

Copyright

   Copyright (C) 2001 Frank J. Tobin, ftobin@neverending.org

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of the
   License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA or see http://www.gnu.org/copyleft/lesser.html
"""

import sys
import os
import os.path
import SocketServer
import socket
import signal
import traceback
import fdpass
import syslog


__author__   = "Frank J. Tobin, ftobin@neverending.org"
# if you update the version here bump the setup.py too!
__version__  = "0.4.0.p3"
__revision__ = "" # Was: $Id: readyexec.py,v 1.28 2002/10/10 01:44:22 ftobin Exp $

stds = ('stdin', 'stdout', 'stderr')

syslog.openlog("postconfirmd", syslog.LOG_PID)
syslog.syslog("Loading '%s' (%s)" % (__name__, __file__))

class ProtocolError(Exception):
    pass

class TimeoutError(Exception):
    pass

class Singleton(object):
    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get('__it__')
        if it is None:
            cls.__it__ = object.__new__(cls)
        return cls.__it__

class Output(Singleton):

    quiet = 0
    do_debug = 0
    #syslog.openlog("sockserver", syslog.LOG_PID)
    def __init__(self, quiet=None, debug=None):
        if quiet is not None: self.quiet = quiet
        if debug is not None: self.do_debug = debug
    def data(self, msg):
        print msg
    def err(self, msg):
        syslog.syslog('%s\n' % msg)
    def warn(self, msg):
        if not self.quiet: syslog.syslog('%s\n' % msg)
    def debug(self, msg):
        if self.do_debug: syslog.syslog('%s\n' % msg)


class ReadyExecHandler(SocketServer.StreamRequestHandler, object):
    """Assumes we are being called in a throw-away environment
    (e.g., a separate process), because we modify
    sys.stdin, sys.stdio, and sys.argv, without saving backups.
    """

    pretend_arg0_fmt = '(readyexec %s)'
    negotiate_secs   = 10
    handler_timeout  = 300

    # We need the rbufsize to be unbuffered because
    # fdpass takes a socket file descriptor, and
    # that screws up the stdio buffering.
    rbufsize         = 0
    
    do_read_args     = True
    do_read_env      = True
    do_read_fd_names = True
    
    def __init__(self, request, client_address, server):
        self.output = Output()
        self.pretend_arg0 = self.pretend_arg0_fmt % self.to_run_path
        super(ReadyExecHandler, self).__init__(request,
                                               client_address, server)

    def handle(self):
        signal.signal(signal.SIGALRM, socket_read_timeout_handler)
        signal.alarm(self.negotiate_secs)
        try:
            op = self.read_string()
        except TimeoutError:
            raise SystemExit(4)

        signal.alarm(0)

        if op == 'conduit':
            self.handle_conduit()
        elif op == 'stop':
            self.handle_stop()
        else:
            raise ProtocolError, "received unknown op %s" % repr(op)

    def handle_conduit(self):
        pid = os.fork()

        if pid == 0:
            try:
                self.handle_conduit_as_subchild()
            except SystemExit, e:
                exit_code = int(str(e))
            except:
                exit_code = 3
                self.output.err(traceback.format_exc().replace('\n','\\n'))
            else:
                exit_code = 0
            os._exit(exit_code)
        
        self.output.debug("waiting on child")
        pid, status = os.waitpid(pid, 0)

        signal.signal(signal.SIGALRM, socket_write_timeout_handler)
        signal.alarm(self.negotiate_secs)
        try:
            self.tell_exit(status >> 8)
        except TimeoutError:
            return

    def handle_conduit_as_subchild(self):
        sys.argv = [self.pretend_arg0]
        signal.signal(signal.SIGALRM, socket_read_timeout_handler)
        signal.alarm(self.negotiate_secs)
        new_std_fds = {}
        
        try:
            if self.do_read_args:
                args = self.read_args()
                self.output.debug("read args %s" % args)
                sys.argv.extend(args)

            if self.do_read_env:
                os.environ = self.read_environ()
            
            for std in stds:
                fd = self.read_fd(std)
                new_std_fds[std] = fd
                self.output.debug("fd for %s is %d" % (std, fd))
            
            self.request.shutdown(0)
        except TimeoutError:
            raise SystemExit(4)
        signal.alarm(0)

        signal.signal(signal.SIGALRM, fd_setup_timeout_handler)
        signal.alarm(self.negotiate_secs)
        try:
            # want to iterate over these in a specific order
            for std in stds:
                std_base = '__%s__' % std
                fd = new_std_fds[std]
                current_file = getattr(sys, std_base)
                if current_file.mode == 'w':
                    current_file.flush()
                current_fno = current_file.fileno()
                os.dup2(fd, current_fno)
                newfile = os.fdopen(current_fno, current_file.mode)
                for s in std_base, std:
                    setattr(sys, s, newfile)
        except TimeoutError:
            raise SystemExit(4)
        signal.alarm(0)
        
        signal.signal(signal.SIGALRM, handler_timeout_handler)
        signal.alarm(self.handler_timeout)
        try:
            apply(self.to_run)
        except TimeoutError:
            raise SystemExit(4)
        finally:
            signal.signal(signal.SIGALRM, flush_timeout_handler)
            signal.alarm(self.negotiate_secs)
            try:
                for std in sys.stdout, sys.stderr:
                    std.flush()
            except TimeoutError:
                raise SystemExit(4)
        signal.alarm(0)


    def handle_stop(self):
        self.stop_server()
    
    def stop_server(self):
        os.kill(self.server.pid, signal.SIGTERM)

    def read_args(self):
        self.output.debug("expecting arguments")
        self.verify_expected("args")
        argc = self.read_long()

        argv = []
        for i in xrange(argc):
            argv.append(self.read_string())

        return argv

    def read_fd(self, stream_name):
        self.output.debug("expecting %s fd" % stream_name)
        if self.do_read_fd_names:
            self.verify_expected(stream_name)
        return fdpass.recv_fd(self.request.fileno())

    def read_environ(self):
        self.output.debug("expecting environment")
        self.verify_expected('env')
        env = {}
        count = self.read_long()

        for i in range(count):
            self.output.debug("reading environment var %d" % (i+1))
            (key, value) = self.read_string().split('=', 1)
            env[key] = value

        return env
    
    def tell_exit(self, code):
        self.send_string("exit")
        self.send_long(code)

    def send_string(self, msg):
        self.output.debug("sending netstring: %s" % repr(msg))
        self.wfile.write(netstring(msg))

    def send_long(self, msg):
        self.output.debug("sending netint: %s" % repr(msg))
        self.wfile.write(netint(msg))

    def read_string(self):
        s = read_netstring(self.rfile)
        self.output.debug("read string %s" % repr(s))
        return s

    def read_long(self):
        l = read_netint(self.rfile)
        self.output.debug("read long %s" % repr(l))
        return l

    def verify_expected(self, expected):
        got = self.read_string()
        if got != expected:
            raise ProtocolError, \
                  "expected %s, got %s" % (repr(expected), repr(got))


class ReadyExec(SocketServer.ForkingMixIn,
                SocketServer.UnixStreamServer, object):
    handled_signals = [signal.SIGINT, signal.SIGTERM]
    
    def __init__(self, to_run, cs_path, quiet=0, debug=0):
        self.output = Output(quiet=quiet, debug=debug)

        assert callable(to_run), ("%s is not callable" % to_run)
        to_run_path = "%s.%s" % (to_run.__module__, to_run.func_name)
        
        ReadyExecHandler.to_run      = staticmethod(to_run)
        ReadyExecHandler.to_run_path = to_run_path

        self.install_signal_handlers()
        
        self.output.debug("listening on %s" % cs_path)
        super(ReadyExec, self).__init__(cs_path, ReadyExecHandler)

    def handle_signal(self, sig, frame):
        syslog.syslog("Terminating.")
        sys.exit()
    handle_signal = classmethod(handle_signal)

    def install_signal_handlers(self):
        for sig in self.handled_signals:
            signal.signal(sig, self.handle_signal)
    install_signal_handlers = classmethod(install_signal_handlers)

    def reset_signal_handlers(self):
        for sig in self.handled_signals:
            signal.signal(sig, signal.SIG_DFL)
    reset_signal_handlers = classmethod(reset_signal_handlers)

    def server_close(self):
        if os.path.exists(self.server_address):
            os.unlink(self.server_address)
        super(ReadyExec, self).server_close()

    def handle_request(self):
        """Overriding so that there is no try statement around
        process_request"""
        try:
            request, client_address = self.get_request()
        except socket.error:
            return

        if self.verify_request(request, client_address):
            self.process_request(request, client_address)
            self.close_request(request)

    def process_request(self, request, address):
        """Fork a new subprocess to process the request.
        Overriding ForkingMixin to not trap exceptions
        """
        # we set the pid here in case we've forked after __init__
        # was called
        if not hasattr(self, 'pid'):
            self.pid = os.getpid()
            
        self.collect_children()
        pid = os.fork()

        if pid == 0:
            # Child process.
            # This must never return, hence os._exit()!
            try:
                self.finish_request(request, address)
            except SystemExit, e:
                exit_code = int(str(e))
            except Exception, e:
                exit_code = 4
                self.output.err(traceback.format_exc().replace('\n','\\n'))
            else:
                exit_code = 0
            os._exit(exit_code)

        # Parent process
        if self.active_children is None:
            self.active_children = []
        self.active_children.append(pid)
        self.close_request(request)

    def finish_request(self, request, address):
        # We reset the signal handlers so that if a child
        # process receives a signal it doesn't affect the head
        # server.
        # Note: this is the first thing that gets executed
        # in the child process.
        self.reset_signal_handlers()
        super(ReadyExec, self).finish_request(request, address)


def netint(i):
    """Returns a netint for i"""
    return "%d," % i

def read_netint(f):
    """Returns the next netint from file f"""
    (l, c) = read_uint(f)
    expect_term = ','
    if c != expect_term:
        raise ProtocolError, \
              "invalid netint termination: expected %c, got %c" \
              % (repr(expect_term), repr(c))
    return long(l)


def netstring(str):
    """Returns a netstring for str
    http://cr.yp.to/proto/netstrings.txt"""
    return "%d:%s," % (len(str), str)


def read_netstring(f):
    """Returns the next netstring read from file f

    http://cr.yp.to/proto/netstrings.txt"""
    (length, c) = read_uint(f)
    expect_term = ':'
    if c != expect_term:
        raise ProtocolError, \
              "invalid netstring length termination: expected %c, got %c" \
              % (repr(expect_term), repr(c))

    
    s = f.read(length)
    expect_term = ','
    c = f.read(1)
    if c != expect_term:
        raise ProtocolError, \
              "invalid netstring termination: expected %c, got %c" \
              % (repr(expect_term), repr(c))

    return s


def read_uint(f, maxdigits=4):
    """This is an ugly hack to deal with the lack of a nice
    stdio-buffering layer on top of sockets so that
    I can read things with scanf() easily

    Note to self: look into recv(2)'s MSG_PEEK
    
    returns (int, terminator)
    """
    n = ''
    i = 0
    
    for i in range(maxdigits+1):
        c = f.read(1)
        if not c.isdigit():
            if not c:
                raise ProtocolError, "received EOF while expecting uint"
            if not n:
                raise ProtocolError, "long cannot start with %s" \
                      % repr(c)
            return (long(n), c)
        n += c

    raise ProtocolError, "%s is getting to be too long" % repr(n)


def socket_read_timeout_handler(signum, frame):
    msg = "Timeout waiting for negotiation on socket"
    syslog.syslog(msg)
    raise TimeoutError(msg)

def socket_write_timeout_handler(signum, frame):
    msg = "Timeout waiting for socket write to complete"
    syslog.syslog(msg)
    raise TimeoutError(msg)

def fd_setup_timeout_handler(signum, frame):
    msg = "Timeout waiting for file descriptor setup"
    syslog.syslog(msg)
    raise TimeoutError(msg)

def handler_timeout_handler(signum, frame):
    msg = "Timeout waiting for handler to complete"
    syslog.syslog(msg)
    raise TimeoutError(msg)

def flush_timeout_handler(signum, frame):
    msg = "Timeout waiting for output streams to be flushed"
    syslog.syslog(msg)
    raise TimeoutError(msg)



# service.py

""" The handler which constitute the actual service.

    Provides whitelist lookups, and generation and verification of
    confirmation requests.  Incoming emails are read on stdin, emails from
    whitelisted or confirmed senders are written to stdout.  Emails are cached
    in a cache directory which is expected to be cleaned out regularly by a
    cronjob.

"""

import os
import re
import sys
import time
import tempfile
import syslog
import base64
import hashlib
import struct
import interpolate
import email
import email.parser
import email.utils
import hmac
import signal
import smtplib
import urllib
import dns.resolver
import dns.rdatatype

from email.MIMEText import MIMEText

#from flufl import bounce
#from datetime import datetime as Datetime, timedelta as Timedelta

log = syslog.syslog

syslog.openlog("postconfirmd", syslog.LOG_PID | syslog.LOG_USER)
log("Loading '%s' (%s)" % (__name__, __file__))

confirm_fmt = "Confirm: %s:%s:%s"
confirm_pat = "Confirm:[ \t\n\r]+.*:.*:.*"
timestamp_fmt = "%Y-%m-%dT%H:%M:%S"

def err(msg):
    sys.stderr.write("Error: %s\n" % (msg, ))
    log(syslog.LOG_ERR, msg)

def debug(expr, locals):
    msg = '%s: %s' % (expr, eval(expr, globals(), locals))
    log(syslog.LOG_INFO, msg)

# ------------------------------------------------------------------------------
def filetext(file):
    if os.path.exists(file):
        file = open(file)
        text = file.read()
        file.close
        return text
    else:
        return ""
        
# ------------------------------------------------------------------------------
# Setup: read initial data before daemonizing.

conf = {}
listinfo = {}
pid  = None
whitelist = set([])
blacklist = set([])
#bouncelist = {}
whiteregex = None
blackregex = None
hashkey  = None

# ------------------------------------------------------------------------------
def read_whitelist(files):
    global whitelist

    whitelist = set([])
    for file in files:
        if os.path.exists(file):
            file = open(file)
            entries = set(file.read().lower().split())
            whitelist |= entries
            file.close()
            log("Read %s whitelist entries from %s\n" % (len(entries), file.name))
    log("Whitelist size: %s" % (len(whitelist)))

# ------------------------------------------------------------------------------
def read_regexes(files):

    regexlist = []
    for file in files:
        if os.path.exists(file):
            file = open(file)
            entries = file.read().split()
            file.close()
            for entry in entries:
                try:
                    re.compile(entry)
                except Exception, e:
                    err("Invalid regex (not added to whitelist): %s\n  Exception: %s" % (entry, e))
                else:
                    if not entry in regexlist:
                        regexlist.append(entry)
            log("Read %s regexlist entries from %s\n" % (len(entries), file.name))
    log("Regexlist size: %s" % (len(regexlist)))
    if regexlist:
        regex = "^(%s)$" % "|".join(regexlist)
    else:
        regex = ""        
    return regex

# ------------------------------------------------------------------------------
def read_blacklist(files):
    global blacklist

    blacklist = set([])
    for file in files:
        if os.path.exists(file):
            file = open(file)
            entries = set(file.read().lower().split())
            blacklist |= entries
            file.close()
            log("Read %s blacklist entries from %s\n" % (len(entries), file.name))
    log("Blacklist size: %s" % (len(blacklist)))

# ------------------------------------------------------------------------------
def read_listinfo():
    global listinfo

    listinfo = {}
    try:
        from Mailman import Utils
        from Mailman import MailList
        for name in Utils.list_names():
            mmlist = MailList.MailList(name, lock=False)
            listinfo[name] = {}
            listinfo[name]['archive'] = bool(mmlist.archive)
        log("Read %s mailman listinfo entries" % len(listinfo))
    except Exception as e:
        log("Exception when reading mailman listinfo: %s" % e)

# ------------------------------------------------------------------------------
# def read_bouncelist(file):
#     global bouncelist
# 
#     if os.path.exists(file):
#         with open(file) as file:
#             entries = file.read().splitlines()
#             tuples = [ e.split(None, 1) for e in entries ]
#             bouncelist = dict([ (a,Datetime.strptime(t, timestamp_fmt)) for t,a in tuples ])
#         log("Read %s bouncelist entries from %s\n" % (len(entries), file.name))
#     log("Bouncelist size: %s" % (len(bouncelist)))

# ------------------------------------------------------------------------------
# def write_bouncelist(bouncelist):
#     cutoff = Datetime.now() - Timedelta(hours=conf.remember_bounce_hours)
#     with open(conf.bouncelist, "w") as file:
#         for a, t in bouncelist.items():
#             if t > cutoff:
#                 file.write("%s %s\n" % (t.strftime(timestamp_fmt), a))

# ------------------------------------------------------------------------------
# def update_bouncelist(addresses):
#     global bouncelist
# 
#     t = Datetime.now()
#     for a in addresses:
#         bouncelist[a.lower()] = t
#     write_bouncelist(bouncelist)
#     # Tell root daemon process to re-read the data files
#     os.kill(pid, signal.SIGHUP)

# ------------------------------------------------------------------------------
def read_data():
    global whiteregex
    global blackregex
    global listinfo
    
    t1 = time.time()
    try:
        read_whitelist(list(conf.whitelists) + [ conf.confirmlist ])
    except Exception as e:
        log("IOError: %s" % e)

    try:
        whiteregex = read_regexes(list(conf.whiteregex))
    except Exception as e:
        log("IOError: %s" % e)

    try:
        read_blacklist(list(conf.blacklists))
    except Exception as e:
        log("IOError: %s" % e)

    try:
        blackregex = read_regexes(list(conf.blackregex))
    except Exception as e:
        log("IOError: %s" % e)

    try:
        read_listinfo()
    except Exception as e:
        log("IOError: %s" % e)

#     try:
#         read_bouncelist(conf.bouncelist)
#     except Exception as e:
#         log("IOError: %s" % e)


    t2 = time.time()
    log(syslog.LOG_INFO, "Wall time for reading data: %.6f s." % (t2 - t1))    

# ------------------------------------------------------------------------------
def sighup_handler(signum, frame):
    """Re-read our data files"""
    read_data()

# ------------------------------------------------------------------------------
def setup(configuration, files):
    global conf
    global pid
    global hashkey

    conf = configuration
    if "mailman_dir" in conf:
        sys.path.append(conf["mailman_dir"])

    pid = os.getpid()

    hashkey = filetext(conf.key_file)

    read_data()

    signal.signal(signal.SIGHUP, sighup_handler)


# ------------------------------------------------------------------------------
def send_smtp(fromaddr, toaddrs, msg, host, port):
    try:
        server = smtplib.SMTP(host, port)
        #server.set_debuglevel(1)    
        server.sendmail(fromaddr, toaddrs, msg)
        #log("SMTP %s:%s  %s -> %s" % (host, port, fromaddr, toaddrs))
        server.quit()
    except Exception as e:
        log(syslog.LOG_ERR, "Error sending mail: %s" % (e, ))
        raise

# ------------------------------------------------------------------------------
def sendmail(sender, recipients, subject, text, host='localhost', port=25, headers={}):

    msg = MIMEText(text)
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = recipients
    for key in headers:
        msg[key] = headers[key]
    message = msg.as_string()

    try:
        send_smtp(sender, recipients, message, host, port)
        #log("Sent mail from '%s' to '%s' about '%s'" % (sender, recipient, subject))
        return True
    except Exception, e:
        log(repr(e))

# ------------------------------------------------------------------------------
def get_msgid(msg):
    msgid = msg['Message-Id']
    if not msgid:
        msgid = msg['Resent-Message-Id']
    if msgid:
        msgid = msgid.strip('<>')
    else:
        msgid = email.utils.make_msgid('ARCHIVE')
    return msgid


# ------------------------------------------------------------------------------
def cache_mail():
    global listinfo

    seconds = base64.urlsafe_b64encode(struct.pack(">i",int(time.time()))).strip("=")
    outfd, outfn = tempfile.mkstemp("", seconds, conf.mail_cache_dir)

    text = sys.stdin.read()
    all = True
    out = os.fdopen(outfd, "w+")
    msg = email.message_from_string(text)
    if conf.archive_url_pattern:
        # The following assumes that the 'List-Id' header field has been set
        # properly in the wrapper before sending the message to postconfirmd:
        if msg['List-Id']:
            from Mailman import MailList # This has to happen after we're configured
            try:
                list = re.search('<([^>]+)>', msg['List-Id']).group(1).replace('.ietf.org', '')
            except Exception:
                list = msg['List-Id'].strip('<>').replace('.ietf.org', '')
            if not list in listinfo:
                try:
                    mmlist = MailList.MailList(list, lock=False)
                    log("Mailman list archive setting for %s: %s" % (list, mmlist.archive))
                except Exception as e:
                    mmlist = None
                    log("No mailman info for list %s: %s" % (list, e))
                listinfo[list] = {}
                listinfo[list]['archive'] = mmlist != None and bool(mmlist.archive)
            if listinfo[list]['archive']:
                msgid = get_msgid(msg)
                sha = hashlib.sha1(msgid)
                sha.update(list)
                hash = base64.urlsafe_b64encode(sha.digest()).strip("=")
                msg["Archived-At"] = conf.archive_url_pattern % {'list': list, 'hash': hash, "msgid": msgid}
                log("Setting Archived-At: %s" % msg["Archived-At"])
            else:
                log("Listinfo[%s]['archive'] = %s" % (list, listinfo[list]['archive']))
        out.write(msg.as_string(unixfrom=True))
    else:
        msg = text
        out.write(text)
    out.close()
    return outfn, msg, all


# ------------------------------------------------------------------------------
def hash(bytes):
    return base64.urlsafe_b64encode(hmac.new(hashkey, bytes, hashlib.sha224).digest()).strip("=")

# ------------------------------------------------------------------------------
def make_hash(sender, recipient, filename):
    return hash( "%s-%s-%s" % (sender, recipient, filename, ) )

# ------------------------------------------------------------------------------
def pad(bytes):
    "Pad with '=' till the string is a multiple of 4"
    return bytes + "=" * ((4 - len(bytes) % 4 ) % 4)

# ------------------------------------------------------------------------------
def request_confirmation(sender, recipient, cachefn, msg):
    """Generate a confirmation request, and send it to the poster for confirmation.

    The request subject line contains:
        'Confirm: <recipient>-<date>-<cachefn>-<hash>'
        Hash is calculated over:
        '<key><sender>-<recipient>-<cachefn><key>'

        <sender> is cleartext
        <recipient> is cleartext
        <cachefn> is the tempfile.mkstmp basename
        <key> is binary
    """
    #log('request_confirmation(%s, %s, %s, ...)' % (sender, recipient, cachefn, ))
    filename = cachefn.split("/")[-1]

    precedence = msg.get("precedence", "")
    precedence_match = re.search(conf.bulk_regex, precedence)
    if precedence_match:
        log(syslog.LOG_INFO, "Skipped confirmation for 'Precedence: %s' message from <%s>" % (precedence, sender,))
        # leave message in cache till cleaned out
        return 1

    body = msg.get_payload()
    while isinstance(body, list):
        body = body[0].get_payload()
    text = body
    confirm_match = re.search(confirm_pat, text)
    if confirm_match:
        dummy, rcp, fn, hash = confirm_match.group(0).rsplit(":", 3)
        if valid_hash(sender, rcp.lstrip(), fn, hash):
            # We have a valid confirmation code inside a message which did
            # not have a confirmation code in the subject line.  It's very
            # likely that this means it's a non-delivery message for an
            # earlier confirmation request.  Don't send a confirmation
            # message for this message, but leave it in cache for possible
            # manual handling
            log(syslog.LOG_INFO, "Skipped confirmation for auto-reply, body contained valid '%s'" % (confirm_match.group(0),))
            return 1

    if sender.lower() in set([ "", recipient.lower() ]):
        log(syslog.LOG_INFO, "Skipped requesting confirmation from <%s>" % (sender,))
        os.unlink(cachefn)
        return 1

    if sender.lower() in blacklist or (blackregex and re.match(blackregex, sender.lower())):
        log(syslog.LOG_INFO, "Skipped confirmation from blacklisted <%s>" % (sender,))
        os.unlink(cachefn)
        return 1

#     if sender.lower() in bouncelist:
#         cutoff = Datetime.now() - Timedelta(hours=conf.remember_bounce_hours)        
#         bouncetime = bouncelist[sender.lower()]
#         if bouncetime > cutoff:
#             log(syslog.LOG_INFO, "Skipped confirmation from bouncing <%s>" % (sender,))
#             os.unlink(cachefn)
#             return 1
#         else:
#             update_bouncelist([])
            
    hash_output = make_hash(sender, recipient, filename)

    log(syslog.LOG_INFO, "Requesting confirmation: <%s> to %s, '%s', %s" % (sender, recipient, filename, hash_output))

    #print "Hash input: %s" % (hash_input, )
    #print "Hash output: %s" % (hash_output, )

    # The ':' is a convenient separator character as it's not permitted in
    # email addresses, and it's not used in our base64 alphabet.  It also
    # conveniently will lets us separate out the Confirm: part at the same
    # time as we split the interesting stuff into parts.
    subject = confirm_fmt % (recipient, filename, hash_output)

    template = filetext(conf.mail_template)
    text = interpolate.interpolate(template, {'filename':filename, 'sender':sender, 'recipient':recipient, 'msg':msg, 'conf':conf})

    sendmail(recipient, sender, subject, text, conf.confirm.smtp.host, conf.confirm.smtp.port)

    return 1
    

# ------------------------------------------------------------------------------
def verify_confirmation(sender, recipient, msg):
    log(syslog.LOG_DEBUG, "Verifying confirmation from <%s> to %s" % (sender, recipient))
    
    subject = msg.get("subject", "")
    subject = re.sub("\s", "", subject)
    dummy, recipient, filename, hash = subject.rsplit(":", 3)

    # Require the sender to be somebody
    if sender == "":
        return 1

    # Require the sender to be different from the recipient
    if sender == recipient:
        return 1

    if not valid_hash(sender, recipient, filename, hash):
        return 1

    # We have a valid confirmation -- update the whitelist and the
    # confirmation file
    if sender.lower() in whitelist:
        log(syslog.LOG_INFO, "Already in whitelist: <%s>" % (sender, ))
    else:
        log(syslog.LOG_INFO, "Adding <%s> to whitelist" % (sender, ))
        whitelist.add(sender)
        file = open(conf.confirmlist, "a")
        file.write("%s\n" % sender)
        file.close()

        # Tell root daemon process to re-read the data files
        os.kill(pid, signal.SIGHUP)

    # Output the cached message and delete the cache file
    log(syslog.LOG_INFO, "Forwarding cached message from <%s> to %s" % (sender, recipient, ))
    cachefn = os.path.join(conf.mail_cache_dir, filename)
    forward_cached_post(cachefn)

    return 0

# ------------------------------------------------------------------------------
def valid_hash(sender, recipient, filename, hash):

    # Require a corresponding message in the cache:
    cachefn = os.path.join(conf.mail_cache_dir, filename)
    if not os.path.exists(cachefn):
        log(syslog.LOG_WARNING, "No cached message for confirmation '%s'" % (filename))
        return False

    # Require that the hash matches
    good_hash = make_hash(sender, recipient, filename)
    if not good_hash == hash:
        log(syslog.LOG_WARNING, "Received hash didn't match -- make_hash(<%s>, '%s', '%s') -> %s != %s" % (sender, recipient, filename, good_hash, hash))
        return False

    return True

# ------------------------------------------------------------------------------
def forward_cached_post(cachefn):
    file = open(cachefn)
    while True:
        text = file.read(8192)
        if not text:
            break
        sys.stdout.write(text)
    file.close()
    os.unlink(cachefn)

# ------------------------------------------------------------------------------
def forward_whitelisted_post(sender, recipient, cachefn, msg, all):
    log(syslog.LOG_DEBUG, "Forwarding from whitelisted <%s> to %s" % (sender, recipient))
    if all:
        sys.stdout.write(msg.as_string())
    else:
        forward_cached_post(cachefn)
    return 0

# ------------------------------------------------------------------------------
def strip_batv(sender):
    if "=" in sender and re.search("^[A-Za-z0-9-]+=[A-Za-z0-9-]+=[^=]+@", sender):
        log(syslog.LOG_INFO, "Stripping BATV prefix from local part of '%s'" % (sender))
        return re.sub("^[A-Za-z0-9-]+=[A-Za-z0-9-]+=", "", sender)
    else:
        return sender

# ------------------------------------------------------------------------------
def parse_options():
    import argparse
    import textwrap

    help = """
    NAME
      postconfirm - Mail confirmation system

    ...  postconfirmd [options]

    DESCRIPTION

      postconfirm consists of postconfirmd, which is the long-running
      (daemon) part, and postconfirmc, which is the client part of a
      client-server program which handles email confirmations. It is
      intended as a front-end to mailing lists. It provides
      funcitonality which is a subset of TMDA, but is adapted to
      high-volume usage and does not have anywhere near all the bells
      and whistles which TMDA has. On the other hand, since the
      whitelist lookup is done by the long-running server part, the
      overhead of doing a verification that a poster has a confirmed
      address is much smaller than for TMDA.

      Early tests indicate that with a whitelist of 16000 entries a lookup
      costs about 0.01 seconds and about ~5 Mbytes for the server daemon, in
      contrast whith TMDA which cost between ~10s and ~600s and ~98 Mb *per
      lookup* as it was set up in front of the IETF lists.

    ACTIONS
      confirm (the default action):

      Somewhat simplified, the flow for regular mail is as follows:

         mail agent
             |
             v
         postconfirm/mailman wrapper [copy] -> cached in cache dir
             |
             v
         postconfirmc <==> postconfirmd
             |
             v
         postconfirm/mailman wrapper [white] -> Mailman
           [grey]
             |
             v
         (waiting for confirmation)

      In the start of the flow above, a message comes in. If the envelope
      From is in the lists known by postconfirmd, it is passed through the
      mailman/ postconfirm wrapper, where postconfirmc is called to talk
      with postconfirmd and get either an OK to pass the message to Mailman,
      or cache it.

      In more detail, this is the sequence of actions that postconfirm goes
      through when processing an incoming email message:

      1. The mail is put in the cache directory.

      2. If the message is a delivery-status report (content-type
         multipart/report, report-type delivery-status) then this is logged,
         and no further action is taken.

      3. If the precedence matches the bulk_regex setting in the
         configuration file, then:

         a. if the envelope sender is in the whitelist or the white_regex
             list, then the message is forwarded directly.

         b. if not, the receipt of a bulk message is logged, and no further
             action is taken.

      4. If the message has an auto-submitted header, and the value matches
         the auto_submitted_regex in the configuration file, then:

         a. if the envelope sender is in the whitelist or the white_regex
            list, then the message is forwarded directly.

         b. if not, the receipt of a bulk message is logged, and no further
            action is taken.

      5. If the subject matches the regex for postconfirm's confirmation
         email subject, the confirmation is processed, and then:

         a. if the confirmation is valid, the held message is forwarded,
            and the confirmed address is added to the whitelist.

         b. if not, the failed confirmation is logged, and the message
            is saved.

      6. If the envelope sender is in the whitelist or the white_regex list,
         the message is forwarded.

      7. If there was no match earlier, a confirmation request is sent out
         for the message.

      dmarc-rewrite:

      For use with mailing lists, aliases, etc., where the presence of
      strict dmarc rules will prevent acceptance of the list messages
      unless rewriting is done.

      dmarc-reverse:

      The inverse operation of dmarc-rewrite. Undoes the rewrite, 
      and replaces the rewritten address with the original address.

            
    OPTIONS...

    FILES
      postconfirmd reads its configuration from the following files (if
      found), in the given order:

          <bin-dir>/postconfirm.conf
          /etc/postconfirm.conf
          /etc/postconfirm/postconfirm.conf
          ~/.postconfirmrc

      where <bin-dir> is the directory where postconfirmd is installed.

    AUTHOR
      Written by Henrik Levkowetz, <henrik@merlot.tools.ietf.org>. Uses the
      daemonize module from Chad J. Schroeder, from the Python Cookbook at
      http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/278731, and
      the readyexec module from Frank J. Tobin, available at
      http://readyexec.sourceforge.net/.

    COPYRIGHT
      Copyright 2008-2017 Henrik Levkowetz

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or (at
      your option) any later version. There is NO WARRANTY; not even the
      implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
      PURPOSE. See the GNU General Public License for more details.

    """
    _prolog, _middle, _epilog = textwrap.dedent(help.expandtabs()).split('...', 2)

    class HelpFormatter(argparse.RawDescriptionHelpFormatter):
        def _format_usage(self, usage, actions, groups, prefix):
            if prefix is None or prefix == 'usage: ':
                prefix = 'SYNOPSIS\n  '
            return _prolog+super(HelpFormatter, self)._format_usage(usage, actions, groups, prefix)

    parser = argparse.ArgumentParser("postconfirmc", description=_middle, epilog=_epilog,
                                     formatter_class=HelpFormatter, add_help=False)

    group = parser.add_argument_group(argparse.SUPPRESS)

    group.add_argument('action', metavar='{ confirm | dmarc-rewrite | dmarc-reverse }',
                                                        help="the action to take")
    group.add_argument('-h', '--help', action='help',   help="show this help message and exit")
    group.add_argument('--stop', action='store_true',   help="stop the postconfirmd daemon")
    group.add_argument('--socket', metavar='SOCKET',    help="use the given SOCKET for client-daemon communication")
    group.add_argument('-f', '--sender',                help="the envelope sender address")
    group.add_argument('-e', '--echo', action='store_true',
                                                        help="just echo back the command line")
    group.add_argument('recipient', nargs='*',          help="recipients, used with dmarc-rewrite and dmarc-reverse commands")

    options = parser.parse_args()

    return options

# ------------------------------------------------------------------------------

def confirm(sender, recipient):
    """ Lookup email sender in whitelist; forward or cache email pending confirmation.

        It is expected that the following environment variables set:
        SENDER   : Envelope MAIL FROM address
	RECIPIENT: Envelope RCPT TO address

        The service handler looks up the sender in the whitelist, and if found,
        the mail on stdin is passed out on stdout.  If not found, the mail is
        instead cached in the configured mail cache directory, and a
        confirmation request is sent to the sender address.  If a reply to the
        confirmation mail is received while the original mail is in the cache,
        the original mail is sent out stdout.  It is expected that a cronjob
        cleans out the cache with desired intervals and cache retention time.
    """

    cachefn, msg, all = cache_mail()

    err = 0
    # With a return code of 0, the incoming message will be forwarded
    # With a return code of 1, the incoming message will be left in cache until it's cleared out.
    # All other return codes indicate postconfirmd failures, and the message will be passed through

    precedence = msg.get("Precedence", "")
    precedence_match = re.search(conf.bulk_regex, precedence)
    auto_submitted = msg.get("Auto-Submitted", "")
    auto_submitted_match = re.search(conf.auto_submitted_regex, auto_submitted)
#    bounced_addresses = bounce.scan_message(msg)

    if (msg.get_content_type() == "multipart/report"
        and ("report-type","delivery-status") in msg.get_params([]) ):
        log(syslog.LOG_INFO, "Received a delivery-status report: %s"  % cachefn)
        err = 1
#     elif bounced_addresses:
#         log(syslog.LOG_INFO, "Received permanent failure delivery status for: %s" % (str(bounced_addresses),))
#         update_bouncelist(bounced_addresses)
#         err = 1
    elif precedence_match:
        if   sender.lower() in whitelist or (whiteregex and re.match(whiteregex, sender.lower())):
            err = forward_whitelisted_post(sender, recipient, cachefn, msg, all)
        else:
            log(syslog.LOG_INFO, "Skipped confirmation for %s message from <%s>" % (precedence, sender,))
            err = 1
        # return code for "don't forward"
        # leaves message in cache till cleaned out        
    elif auto_submitted_match:
        if   sender.lower() in whitelist or (whiteregex and re.match(whiteregex, sender.lower())):
            err = forward_whitelisted_post(sender, recipient, cachefn, msg, all)
        else:
            log(syslog.LOG_INFO, "Skipped confirmation for %s message from <%s>" % (auto_submitted, sender,))
            err = 1
    elif re.search(confirm_pat, msg.get("subject", "")):
        err =  verify_confirmation(sender, recipient, msg)
        if err:
            log(syslog.LOG_INFO, "Message with failed confirmation saved as %s" % (cachefn))
        else:
            os.unlink(cachefn)
    else:
        if   sender.lower() in whitelist or (whiteregex and re.match(whiteregex, sender.lower())):
            err = forward_whitelisted_post(sender, recipient, cachefn, msg, all)
        else:
            err = request_confirmation(sender, recipient, cachefn, msg)

    if err:
        raise SystemExit(err)

# ----------------------------------------------------------------------
# DMARC related
#----------------------------------------------------------------------

def get_dns_record(dom, rtype):
    resolver = dns.resolver.Resolver()
    resolver.lifetime = conf.dmarc.resolver.lifetime
    resolver.timeout = conf.dmarc.resolver.timeout
    try:
        recs = resolver.query(dom, rtype)
    except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
        return None
    except dns.exception.DNSException as error:
        log('DNSException: Unable to resolve %s: %s', dom, error.__doc__)
        return None
    return recs


def dmarc_reject_or_quarantine(domain, org=False):
    # This takes a domain and a flag stating whether
    # we should check the organizational domains.  It returns one of three
    # values:
    # * True if the DMARC policy is reject or quarantine;
    # * False if is not;
    # * A special sentinel if we should continue looking
    dmarc_domain = '_dmarc.%s'%domain
    policy = None
    rrecs = get_dns_record(dmarc_domain, dns.rdatatype.TXT)
    if rrecs == None:
        return False
    # Be as robust as possible in parsing the result.
    results_by_name = {}
    cnames = {}
    # Check all the records returned by DNS.  Keep track of the CNAMEs for
    # checking later on.  Ignore any other non-TXT records.
    for rec in rrecs.response.answer:
        if rec.rdtype == dns.rdatatype.CNAME:
            cnames[rec.name.to_text()] = (rec.items[0].target.to_text())
        if rec.rdtype != dns.rdatatype.TXT:
            continue
        result = ''.join([ record for record in rec.items[0].strings ])
        name = rec.name.to_text()
        results_by_name.setdefault(name, []).append(result)
    #
    want_names = set([dmarc_domain + '.'])
    expands = list(want_names)
    seen = set(expands)
    while expands:
        item = expands.pop(0)
        if item in cnames:
            if cnames[item] in seen:
                # CNAME loop.
                continue
            expands.append(cnames[item])
            seen.add(cnames[item])
            want_names.add(cnames[item])
            want_names.discard(item)
    assert len(want_names) == 1, ('Error in CNAME processing for %s; want_names != 1.'%(dmarc_domain,))
    for name in want_names:
        if name not in results_by_name:
            continue
        dmarcs = [ record for record in results_by_name[name] if record.startswith('v=DMARC1;') ]
        if len(dmarcs) == 0:
            sys.stderr.write("no DMARC1 records for %s\n" % (dmarc_domain, ))
            return None
        if len(dmarcs) > 1:
            log('RRset of TXT records for %s has %d v=DMARC1 entries; ignoring them per RFC 7489, Section 6.6.3 .' %
                (dmarc_domain, len(dmarcs)))
            return None
        for entry in dmarcs:
            match = re.search(r'\bsp=(\w*)\b', entry, re.IGNORECASE)
            if org and match:
                policy = match.group(1).lower()
            else:
                match = re.search(r'\bp=(\w*)\b', entry, re.IGNORECASE)
                if match:
                    policy = match.group(1).lower()
                else:
                    continue
            if policy in ('reject', 'quarantine'):
                return True
    return False
    

def quote(s):
    return urllib.quote(s).replace('%', conf.dmarc.rewrite.quote_char)

def unquote(q):
    return urllib.unquote(q.replace(conf.dmarc.rewrite.quote_char, '%'))

def dmarc_rewrite(sender, recipients):
    text = sys.stdin.read()
    rewrite_done = False
    msg = email.message_from_string(text)
    if conf.dmarc.rewrite.require and conf.dmarc.rewrite.require.header:
        for key in conf.dmarc.rewrite.require.header:
            if msg[key] in conf.dmarc.rewrite.require.header[key]:
                break
        else:
            send_smtp(sender, recipients, text, conf.dmarc.rewrite.smtp.host, conf.dmarc.rewrite.smtp.port)
            #log("No dmarc rewrite requirement matched for %s -> %s" % (sender, recipients))
            return
    from_field = msg["From"]
    from_name, from_addr = email.utils.parseaddr(from_field)
    if from_addr:
        from_local, at, from_dom = from_addr.partition('@')
        if at == '@' and dmarc_reject_or_quarantine(from_dom):
            new_addr = '%s@%s' % (quote(from_addr), conf.dmarc.domain)
            new_from = email.utils.formataddr((from_name, new_addr))
            # do rewrite
            msg['X-Original-From'] = msg['From']
            del msg['From']
            msg['From'] = new_from
            rewrite_done = True

    if rewrite_done:
        send_smtp(sender, recipients, msg.as_string(), conf.dmarc.rewrite.smtp.host, conf.dmarc.rewrite.smtp.port)
        log("Dmarc rewrite: '%s' to '%s' --> %s" % (from_field, new_from, recipients))
    else:
        send_smtp(sender, recipients, text, conf.dmarc.rewrite.smtp.host, conf.dmarc.rewrite.smtp.port)
        #log("No dmarc rewrite done for '%s' --> %s" % (sender, recipients))

# ------------------------------------------------------------------------------

def dmarc_reverse(sender, recipient):
    recipients = []

    text = sys.stdin.read()
    msg = email.message_from_string(text)
    if sender.lower() in whitelist or (whiteregex and re.match(whiteregex, sender.lower())):
        for field in [ 'To', 'Cc', ]:
            addresses = msg.get_all(field)
            if addresses:
                reversed = []
                dirty = False
                for name, addr in email.utils.getaddresses(addresses):
                    oaddr = addr
                    if addr.endswith('@%s'%conf.dmarc.domain):
                        localpart, dom  = addr.rsplit('@', 1)
                        addr = unquote(localpart)
                        recipients.append( email.utils.formataddr((name, addr)) )
                        dirty = True
                        log("dmarc-reverse message-id=%s from=<%s> x-original-to=<%s> to=<%s>" % (msg['message-id'], sender, oaddr, addr, ))
                    reversed.append( email.utils.formataddr((name, addr)) )
                if dirty:
                    del msg[field]
                    for a in reversed:
                        msg[field] = a
        send_smtp(sender, recipients, msg.as_string(), conf.dmarc.reverse.smtp.host, conf.dmarc.reverse.smtp.port)
        return 0
    else:
        log("Received a dmarc-rewrite message from a sender not in whitelist or confirmed list: %s" % sender)
        return 4


# ------------------------------------------------------------------------------
# Service handler

def handler():
    try:
        options = parse_options()

        if options.echo:
            log('sys.argv: %s' % sys.argv)

        if not options.sender:
            if "SENDER" in os.environ:
                options.sender = os.environ["SENDER"]
            else:
                log(syslog.LOG_ERR, "sender not provided -- can't process input")
                return 3

        if not options.recipient:
            if "RECIPIENT" in os.environ:
                options.recipient = os.environ["RECIPIENT"]
            else:
                log(syslog.LOG_ERR, "recipient not provided -- can't process input")
                return 3

        t1 = time.time()

        sender    = strip_batv(options.sender)
        recipient = options.recipient
        if isinstance(recipient, list):
            if options.action != 'dmarc-rewrite':
                if len(recipient) > 1:
                    log(syslog.LOG_ERR, "Only the dmarc-rewrite action accepts multiple recipients: %s" % recipients)
                recipient = recipient[0].strip()
            else:
                recipient = [ r.strip() for r in recipient ]
        else:
            if options.action == 'dmarc-rewrite':
                recipient = [ recipient.strip() ]
            else:
                recipient = recipient.strip()

        if   options.action == 'confirm':
            result = confirm(sender, recipient)
        elif options.action == 'dmarc-rewrite':
            result = dmarc_rewrite(sender, recipient)
        elif options.action == 'dmarc-reverse':
            result = dmarc_reverse(sender, recipient)
        else:
            log(syslog.LOG_ERR, "Bad command: %s" % (options.action, ))
            result = 3

        t2 = time.time()
        log(syslog.LOG_INFO, "Wall time in %s handler: %.6f s." % (options.action, t2 - t1))

        return result

    except Exception as e:
        sys.stderr.write("Exception: %s\n" % e)
        raise
        return 2



"""String interpolation for Python (by Ka-Ping Yee, 14 Feb 2000).

Modified by Henrik Levkowetz 28 Aug 2005, as follows:
    * Added a few doctest samples, invoked when module is run standalone
    * Changed syntax for invocation of arbitrary python expressions from
      ${arbitrary expression} to $[arbitrary expression]
    * Added shell invocation with $(arbitrary shell command)
    * Changed names of several of the functions to be (maybe?) more readable:
        - Iplt -> Interpolator
        - iplt -> interpolate
        - printpl -> iprint

This module lets you quickly and conveniently interpolate values into
strings (in the flavour of Perl or Tcl, but with less extraneous
punctuation).  You get a bit more power than in the other languages,
because this module allows subscripting, slicing, function calls,
attribute lookup, or arbitrary expressions.  Variables and expressions
are evaluated in the namespace of the caller.

The interpolate() function returns the result of interpolating a string, and
iprint() prints out an interpolated string.  Here are some examples:

    from interpolator import iprint
    iprint("Here is a $string.")
    iprint("Here is a $module.member.")
    iprint("Here is an $object.member.")
    iprint("Here is a $functioncall(with, arguments).")
    iprint("Here is an $[arbitrary + expression].")
    iprint("Here is an $array[3] member.")
    iprint("Here is a $dictionary['member'].")
    iprint("Here is an $(echo 'shell command' | tr "s" "S").")

    >>> from interpolate import iprint
    >>> string = "String"
    >>> iprint("Here is a $string.")
    Here is a String.
    >>> iprint("Here is module member $os.listdir.")
    Here is module member <built-in function listdir>.
    >>> object=Interpolator("")
    >>> iprint("Here is object member $object.__init__.")
    Here is object member <bound method Interpolator.__init__ of <Interpolator ''>>.
    >>> iprint("Here is the result of a funciton call: $len('Hello world').")
    Here is the result of a funciton call: 11.
    >>> iprint("Here is arbitrary expression $[1 + 2].")
    Here is arbitrary expression 3.
    >>> array = [0, "1", 2, "Yup"]
    >>> iprint("Here is array member $array[3].")
    Here is array member Yup.
    >>> dictionary = {'hello': 1, 'member': "dictionary member"}
    >>> iprint("Here is a $dictionary['member'].")
    Here is a dictionary member.
    >>> iprint("Here is a $(echo -n 'phell command' | tr 'p' 's').")
    Here is a shell command.


The filter() function filters a file object so that output through it
is interpolated.  This lets you produce the illusion that Python knows
how to do interpolation:

    >>> import interpolate
    >>> sys.stdout = interpolate.filter()
    >>> f = "fancy"
    >>> print "Isn't this $f?"
    Isn't this fancy?
    >>> print "Standard output has been replaced with a $sys.stdout object." #doctest: +ELLIPSIS
    Standard output has been replaced with a <interpolated <doctest._SpoofOut instance at 0x...>> object.
    >>> sys.stdout = interpolate.unfilter()
    >>> print "Okay, back $to $normal."
    Okay, back $to $normal.

Under the hood, the Interpolator class represents a string that knows how to
interpolate values.  An instance of the class parses the string once
upon initialization; the evaluation and substitution can then be done
each time the instance is evaluated with str(instance).  For example:

    >>> from interpolate import Interpolator
    >>> s = Interpolator("Here is $foo.")
    >>> foo = 5
    >>> print str(s)
    Here is 5.
    >>> foo = "bar"
    >>> print str(s)
    Here is bar.
    
"""

import sys, string, os
from types import StringType
from tokenize import tokenprog
import shlex

import syslog
syslog.openlog("postconfirmd", syslog.LOG_PID)
syslog.syslog("Loading '%s' (%s)" % (__name__, __file__))

class InterpolatorError(ValueError):
    def __init__(self, text, pos):
        self.text = text
        self.pos = pos
    def __str__(self):
        return "unfinished expression in %s at char %d" % (
            repr(self.text), self.pos)

def pymatchorfail(text, pos):
    match = tokenprog.match(text, pos)
    if match is None:
        raise InterpolatorError(text, pos)
    return match, match.end()

def shmatchorfail(text, pos):
    lex = shlex.shlex(text[pos:], "'%s...'" % text[pos:pos+6])
    while text[pos] in lex.whitespace:
        pos = pos+1
    match = lex.get_token()
    if match is None:
        raise InterpolatorError(text, pos)
    return match, pos+len(match)

class Interpolator:
    """Class representing a string with interpolation abilities.
    
    Upon creation, an instance works out what parts of the format
    string are literal and what parts need to be evaluated.  The
    evaluation and substitution happens in the namespace of the
    caller when str(instance) is called."""

    def __init__(self, format, loc=None, glob=None):
        """The single argument to this constructor is a format string.

        The format string is parsed according to the following rules:

        1.  A dollar sign and a name, possibly followed by any of: 
              - an open-paren, and anything up to the matching paren 
              - an open-bracket, and anything up to the matching bracket 
              - a period and a name 
            any number of times, is evaluated as a Python expression.

        2.  A dollar sign immediately followed by an open-paren, and
            anything up to the matching close-paren, is evaluated as
            a Python expression.

        3.  Outside of the expressions described in the above two rules,
            two dollar signs in a row give you one literal dollar sign."""

        if type(format) != StringType:
            raise TypeError, "needs string initializer"
        self.format = format
        self.loc = loc
        self.glob = glob
        
        namechars = "abcdefghijklmnopqrstuvwxyz" \
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
        chunks = []
        pos = 0

        while 1:
            dollar = string.find(format, "$", pos)
            if dollar < 0: break
            if dollar > 0:
                prevchar = format[dollar-1]
            else:
                prevchar = ""
            nextchar = format[dollar+1]

            if prevchar == "\\":
                chunks.append((0, format[pos:dollar]))
                pos = dollar + 1

            elif nextchar == "[":
                chunks.append((0, format[pos:dollar]))
                pos, level = dollar+2, 1
                while level:
                    match, pos = pymatchorfail(format, pos)
                    tstart, tend = match.regs[3]
                    token = format[tstart:tend]
                    if token == "[": level = level+1
                    elif token == "]": level = level-1
                chunks.append((1, format[dollar+2:pos-1]))

            elif nextchar == "(":
                chunks.append((0, format[pos:dollar]))
                pos, level = dollar+2, 1
                while level:
                    token, pos = shmatchorfail(format, pos)
                    if token == "(": level = level+1
                    elif token == ")": level = level-1
                chunks.append((2, format[dollar+2:pos-1]))

            elif nextchar in namechars:
                chunks.append((0, format[pos:dollar]))
                match, pos = pymatchorfail(format, dollar+1)
                while pos < len(format):
                    if format[pos] == "." and \
                        pos+1 < len(format) and format[pos+1] in namechars:
                        match, pos = pymatchorfail(format, pos+1)
                    elif format[pos] in "([":
                        pos, level = pos+1, 1
                        while level:
                            match, pos = pymatchorfail(format, pos)
                            tstart, tend = match.regs[3]
                            token = format[tstart:tend]
                            if token[0] in "([": level = level+1
                            elif token[0] in ")]": level = level-1
                    else: break
                chunks.append((1, format[dollar+1:pos]))

            else:
                chunks.append((0, format[pos:dollar+1]))
                pos = dollar + 1

        if pos < len(format): chunks.append((0, format[pos:]))
        self.chunks = chunks

    def __repr__(self):
        return "<Interpolator " + repr(self.format) + ">"

    def __str__(self):
        """Evaluate and substitute the appropriate parts of the string."""
        try: 1/0
        except: frame = sys.exc_traceback.tb_frame

        while frame.f_globals["__name__"] == __name__: frame = frame.f_back
        if self.loc:
            loc = self.loc
        else:
            loc = frame.f_locals
        if self.glob:
            glob = self.glob
        else:
            glob = frame.f_globals

        result = []
        for live, chunk in self.chunks:
            if live == 1:
                result.append(str(eval(chunk, loc, glob)))
            elif live == 2:
                for key in glob.keys():
                    if type(glob[key]) == type(""):
                        os.environ[key] = glob[key]
                for key in loc.keys():
                    if type(loc[key]) == type(""):
                        os.environ[key] = loc[key]
                pipe = os.popen(chunk, "r")
                result.append(pipe.read().rstrip())
                pipe.close()
            else:
                result.append(chunk)

        return string.join(result, "")

def interpolate(text, loc=None, glob=None): return str(Interpolator(text, loc, glob))
def iprint(text): print interpolate(text)

class InterpolatorFile:
    """A file object that filters each write() through an interpolator."""
    def __init__(self, file): self.file = file
    def __repr__(self): return "<interpolated " + repr(self.file) + ">"
    def __getattr__(self, attr): return getattr(self.file, attr)
    def write(self, text): self.file.write(str(Interpolator(text)))

def filter(file=sys.stdout):
    """Return an InterpolatorFile that filters writes to the given file object.
    
    'file = filter(file)' replaces 'file' with a filtered object that
    has a write() method.  When called with no argument, this creates
    a filter to sys.stdout."""
    return InterpolatorFile(file)

def unfilter(ifile=None):
    """Return the original file that corresponds to the given InterpolatorFile.
    
    'file = unfilter(file)' undoes the effect of 'file = filter(file)'.
    'sys.stdout = unfilter()' undoes the effect of 'sys.stdout = filter()'."""
    return ifile and ifile.file or sys.stdout.file

def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()




import os
import smtplib
import syslog
from email.MIMEText import MIMEText

log = syslog.syslog

def sendmail(sender, recipient, subject, text, conf={"smtp_host":"localhost",}, headers={}):

    msg = MIMEText(text)
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = recipient
    for key in headers:
        msg[key] = headers[key]
    message = msg.as_string()

    try:
        server = smtplib.SMTP(conf["smtp_host"])
        server.sendmail(sender, recipient, message)
        server.quit()
        #log("Sent mail from '%s' to '%s' about '%s'" % (sender, recipient, subject))
        return True
    except Exception, e:
        msg["To"] = "webmaster@" + os.environ.get("SERVER_NAME", "localhost")
        msg["Subject"] = "Failed send: %s" % repr(e)
        msg["X-Orig-Subject"] = subject
        msg["X-Orig-To"] = recipient    
        try:
            server = smtplib.SMTP("localhost")
            server.sendmail(sender, recipient, message)
            server.quit()
        except:
            log(repr(e))
        raise e



#!/usr/bin/python
# -*- python -*-

"""
NAME
	%(program)s - Mail confirmation system (daemon part)

SYNOPSIS
	%(program)s [OPTIONS]

DESCRIPTION
        %(program)s is the long-running (daemon) part of a client-server
        program which handles email confirmations. It is intended as a
        front-end to mailing lists. It provides funcitonality which is a
        subset of TMDA, but is adapted to high-volume usage and does not have
        anywhere near all the bells and whistles which TMDA has. On the other
        hand, since the whitelist lookup is done by the long-running server
        part, the overhead of doing a verification that a poster has a
        confirmed address is much smaller than for TMDA.

        Early tests indicate that with a whitelist of 16000 entries a lookup
        costs about 0.01 seconds and about ~5 Mbytes for the server daemon, in
        contrast whith TMDA which cost between ~10s and ~600s and ~98 Mb *per
        lookup* as it was set up in front of the IETF lists.

        Somewhat simplified, the flow for regular mail is as follows:

           mail agent
               |
               v
           postconfirm/mailman wrapper [copy] -> cached in cache directory
               |
               v
           postconfirmc <==> postconfirmd
               |
               v
           postconfirm/mailman wrapper [white] -> Mailman
             [grey]
               |
               v
           (waiting for confirmation)

        In the start of the flow above, a message comes in. If the envelope
        From is in the lists known by postconfirmd, it is passed through the
        mailman/ postconfirm wrapper, where postconfirmc is called to talk
        with postconfirmd and get either an OK to pass the message to Mailman,
        or cache it.
        
        In more detail, this is the sequence of actions that postconfirm goes
        through when processing an incoming email message:

        1. The mail is put in the cache directory.

        2. If the message is a delivery-status report (content-type
           multipart/report, report-type delivery-status) then this is logged,
           and no further action is taken.

        3. If the precedence matches the bulk_regex setting in the
           configuration file, then:
           
           a. if the envelope sender is in the whitelist or the white_regex
               list, then the message is forwarded directly.

           b. if not, the receipt of a bulk message is logged, and no further
               action is taken.

        4. If the message has an auto-submitted header, and the value matches
           the auto_submitted_regex in the configuration file, then:
           
           a. if the envelope sender is in the whitelist or the white_regex
              list, then the message is forwarded directly.

           b. if not, the receipt of a bulk message is logged, and no further
              action is taken.

        5. If the subject matches the regex for postconfirm's confirmation
           email subject, the confirmation is processed, and then:
           
           a. if the confirmation is valid, the held message is forwarded,
              and the confirmed address is added to the whitelist.

           b. if not, the failed confirmation is logged, and the message
              is saved.

        6. If the envelope sender is in the whitelist or the white_regex list,
           the message is forwarded.

        7. If there was no match earlier, a confirmation request is sent out
           for the message.


%(options)s

FILES
        %(program)s reads its configuration from the following files (if found),
        in the given order:
        
            <bin-dir>/postconfirm.conf
            /etc/postconfirm.conf
            /etc/postconfirm/postconfirm.conf
            ~/.postconfirmrc

        where <bin-dir> is the directory where postconfirmd is installed.

AUTHOR
	Written by Henrik Levkowetz, <henrik@merlot.tools.ietf.org>. Uses the
        daemonize module from Chad J. Schroeder, from the Python Cookbook at
        http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/278731, and
        the readyexec module from Frank J. Tobin, available at
        http://readyexec.sourceforge.net/.

COPYRIGHT
	Copyright 2008 Henrik Levkowetz

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or (at
	your option) any later version. There is NO WARRANTY; not even the
	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE. See the GNU General Public License for more details.

"""

import re
import sys
import os.path
import stat
import getopt
import config
import StringIO
import syslog
from __init__ import __version__

# ------------------------------------------------------------------------------
# Misc. metadata


program = os.path.basename(sys.argv[0])
progdir = os.path.dirname(sys.argv[0])

# ------------------------------------------------------------------------------
# Set up logging

syslog.openlog("postconfirmd", syslog.LOG_PID)
log = syslog.syslog

log("Postconfirm daemon v%s starting (%s)." % (__version__, __file__))

# ------------------------------------------------------------------------------
# Read config file


merger = config.ConfigMerger(lambda x, y, z: "overwrite")

default = """
foreground:         False
debug:              False
socket_path:        "/var/run/postconfirm/socket"
archive_url_pattern: "http://mailarchive.ietf.org/arch/msg/%(list)s/%(hash)s"
auto_submitted_regex:	"^auto-"
mailman_dir:        "/usr/lib/mailman"
remember_bounce_hours: 12
confirm:
{
    smtp:
    {
        host: localhost
        port: 25
    }
}
dmarc:
{
    domain:         "dmarc.ietf.org"
    resolver:
    {
        # seconds
        timeout: 3                      
        lifetime: 5
    }
    rewrite:
    {
        quote_char: "="
        smtp:
        {
            host: localhost
            port: 10028
        }
    }
    reverse:
    {
        smtp:
        {
            host: localhost
            port: 10025
        }
    }
}
"""

default = StringIO.StringIO(default)
conf = config.Config(default)

for conffile in [progdir+"/postconfirm.conf",
                "/etc/postconfirm.conf",
                "/etc/postconfirm/postconfirm.conf",
                os.path.expanduser("~/.postconfirmrc"),
            ]:
    if os.path.exists(conffile):
        log("Reading configuration from %s" % conffile)
        merger.merge(conf, config.Config(conffile))
try:
    conf.dmarc.rewrite.quote_char
except AttributeError:
    conf.dmarc.rewrite.quote_char = '='

# import pprint
# pprint.pprint(dict(conf), sys.stdout, 4, 40)

# ------------------------------------------------------------------------------
# Create list of options, for the help text

options = ""
for line in re.findall("\n +(if|elif) +opt in \[(.+)\]:\s+#(.+)\n", open(__file__.replace(".pyc", ".py")).read()):
    if not options:
        options += "OPTIONS\n"
    options += "        %-16s %s\n" % (line[1].replace('"', ''), line[2])
options = options.strip()

# ------------------------------------------------------------------------------
# Process options

# with ' < 1:' on the next line, this is a no-op:
if len(sys.argv) < 1:
    print __doc__ % locals()
    sys.exit(1)

try:
    opts, args = getopt.gnu_getopt(sys.argv[1:], "dfhs:V", ["debug", "foreground", "help", "socket=", "version", ])
except Exception, e:
    print "%s: %s" % (program, e)
    sys.exit(1)

# process option switches
for opt, value in opts:
    if not opt:
        pass
    elif opt in ["-d", "--debug"]:      # Run with debug output
        conf.debug = True
    elif opt in ["-f", "--foreground"]: # Run in the foreground, don't daemonize
        conf.foreground = True
    elif   opt in ["-h", "--help"]:     # Output this help, then exit
        print __doc__ % globals()
        sys.exit(1)
    elif opt in ["-s", "--socket"]:     # Specify the socket path
        conf.socket_path = value
    elif opt in ["-V", "--version"]:    # Output version, then exit
        print program, __version__
        sys.exit(0)


# ------------------------------------------------------------------------------

import daemonize
import grp
import pwd
import service
import sockserver
import syslog

# ------------------------------------------------------------------------------
# Utility functions

def mkdir(path):
    if not os.path.exists(path):
        print "Creating directory '%s'" % path
        os.mkdir(path)
        return True
    return False

def mkfile(file):
    path = os.path.dirname(file)
    mkdir(path)
    if not os.path.exists(file):
        os.mknod(file)
        return True
    return False

# ------------------------------------------------------------------------------
# Some assertions
assert os.path.exists(conf.mail_template), "Could not find mail template at configured path: %s" % conf.mail_template
assert type(conf.whitelists) == config.Sequence, "The whitelist configuration should be a list of file paths."

# ------------------------------------------------------------------------------
# Create directories and do initialization, as needed:

def run():
    sys.stderr.write("\nPostconfirm daemon v%s starting (%s).\n" % (__version__, __file__))

    mkdir(conf.mail_cache_dir)

    if mkfile(conf.key_file):
        keyfile = open(conf.key_file, "w")
        keyfile.write(os.urandom(128))
        keyfile.close()

    mkfile(conf.confirmlist)
    #mkfile(conf.bouncelist)

    mkdir(os.path.dirname(conf.socket_path))



    # ------------------------------------------------------------------------------
    # Maybe daemonize

    # Get user and group we should execute as
    user, pw, uid, gid, gecos, home, shell = list(pwd.getpwnam(conf.daemon_user))
    if "daemon_group" in conf:
        group, gpw, gid, members = list(grp.getgrnam(conf.daemon_group))

    if not conf.foreground:
        pidfname = "/var/run/%s.pid" % program

        pidfile = open(pidfname, "w")
        os.chown(pidfile.name, uid, gid)
        pidfile.close()

        os.setgid(gid)
        os.setuid(uid)
        daemonize.createDaemon()

        pidfile = open(pidfname, "w")
        pidfile.write("%s" % os.getpid())
        pidfile.close()
    else:
        log("Not daemonizing.")
        sys.stderr.write("Not daemonizing.\n")

    sys.stderr.write("Will listen on unix domain socket '%s'\n" % conf.socket_path)

    # ------------------------------------------------------------------------------
    # connect stdout and stderr to syslog, to catch messages, exceptions, traceback
    if not conf.foreground:
        log("Redirecting stdout and stderr to syslog")
        syslog.write = syslog.syslog
        sys.stdout = syslog
        sys.stderr = syslog

    # ------------------------------------------------------------------------------
    # Set up the service
    service.setup(conf, args)

    # ------------------------------------------------------------------------------
    # Start the server

    try:
        server = sockserver.ReadyExec(service.handler, conf.socket_path, debug=conf.debug)
    except IOError as e:
        log(" ** Tried to set up server to read from socket %s, but got an exception: '%s'" % (conf.socket_path, e))
        raise

    os.chmod(conf.socket_path, stat.S_IRWXU | stat.S_IRWXG )
    os.chown(conf.socket_path, uid, gid)

    try:        
        server.serve_forever()
    finally:
        server.server_close()

if __name__ == '__main__':
    run()
    



#!/usr/bin/env python
# --------------------------------------------------
# Copyright The IETF Trust 2011, All Rights Reserved
# --------------------------------------------------

import re
import os.path
import sys

from setuptools import setup, find_packages, Extension
from setuptools.command.install import install
from distutils.ccompiler import new_compiler
from codecs import open

CONF_FILE_DIR = '/etc/postconfirm/'

def maybe_chown(path, uid, gid):
    try:
        os.chown(path, uid, gid)
    except OSError:
        sys.stderr.write("   WARNING: Failed chown %s:%s %s\n" % (uid, gid, path))
        pass

def maybe_chmod(path, mode):
    try:
        os.chmod(path, mode)
    except OSError:
        sys.stderr.write("   WARNING: Failed chmod %s %s\n" % (mode, path))
        pass

class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        import config, pwd, grp, sys, stat

        # -------------------------
        # Run the base installation
        install.run(self)

        # -------------------------
        # Post-install actions
        # Get config
        merger = config.ConfigMerger(lambda x, y, z: "overwrite")
        conf = config.Config()
        conf_file = os.path.join(CONF_FILE_DIR, 'postconfirm.conf')
        for cf in [
                    "/etc/postconfirm.conf",
                    "/etc/postconfirm/postconfirm.conf",
                    conf_file,
                    os.path.expanduser("~/.postconfirmrc"),
                ]:
            if os.path.exists(cf):
                merger.merge(conf, config.Config(cf))
        # Get user and group we should execute as
        user, pw, uid, gid, gecos, home, shell = list(pwd.getpwnam(conf.daemon_user))
        if "daemon_group" in conf:
            group, gpw, gid, members = list(grp.getgrnam(conf.daemon_group))
        # Set correct file owner, group and mode
        maybe_chown(conf_file, uid, gid)
        if not os.path.exists(conf.key_file):
            with open(conf.key_file, "wb") as k:
                with open('/dev/random') as r:
                    rand = r.read(128)
                    k.write(rand)
        maybe_chown(conf.key_file, uid, gid)
        maybe_chmod(conf.key_file, 0o600)
        postconfirmc = os.path.join(self.install_scripts, 'postconfirmc')
        maybe_chown(postconfirmc, uid, gid)
        maybe_chmod(postconfirmc, stat.S_ISUID|0o755 )

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as file:
    long_description = file.read()

# Get the requirements from the local requirements.txt file
with open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as file:
    requirements = file.read().splitlines()

# Get the requirements from the local manifest file
with open(os.path.join(here, 'MANIFEST.in'), encoding='utf-8') as file:
    extra_files = [ l.split()[1] for l in file.read().splitlines() if l ]

conf_files = []
for fn in ['postconfirm.conf', 'confirm.email.template']:
    if not os.path.exists(os.path.join(CONF_FILE_DIR, fn)):
        # this will be added to setup(data_files=...) below, so use data_files
        # semantics:
        sys.stderr.write("Adding conf_file %s\n" % os.path.join(CONF_FILE_DIR, fn))
        conf_files.append( (CONF_FILE_DIR, [ fn, ]) )

def parse(changelog):
    ver_line = "^([a-z0-9+-]+) \(([^)]+)\)(.*?) *$"
    sig_line = "^ -- ([^<]+) <([^>]+)>  (.*?) *$"
    #
    entries = []
    if type(changelog) == type(''):
        changelog = open(changelog)
    for line in changelog:
        if re.match(ver_line, line):
            package, version, rest = re.match(ver_line, line).groups()
            entry = {}
            entry["package"] = package
            entry["version"] = version
            entry["logentry"] = ""
        elif re.match(sig_line, line):
            author, email, date = re.match(sig_line, line).groups()
            entry["author"] = author
            entry["email"] = email
            entry["datetime"] = date
            entry["date"] = " ".join(date.split()[:3])
            entries += [ entry ]
        else:
            entry["logentry"] += line
    changelog.close()
    entry["logentry"] = entry["logentry"].strip()
    return entries

changelog_entry_template = """
Version %(version)s (%(date)s)
------------------------------------------------

%(logentry)s"""

long_description += """
Changelog
=========
""" + "\n".join([ changelog_entry_template % entry for entry in parse("changelog")[:3] ])



import postconfirm

compiler = new_compiler()
client_o_files = compiler.compile(['postconfirmc.c',])
compiler.link_executable(client_o_files, 'postconfirmc')

setup(
    name='postconfirm',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=postconfirm.__version__,

    description='Mailing list posting confirmation daemon',
    long_description=long_description,

    # The project's main homepage.
    #url = ''

    # Author details
    author='Henrik Levkowetz',
    author_email='henrik@levkowetz.com',

    # Choose your license
    license='Simplified BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Other Audience',
        'Topic :: Communications :: Email',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        #'Programming Language :: Python :: 3',
        #'Programming Language :: Python :: 3.3',
        #'Programming Language :: Python :: 3.4',
        #'Programming Language :: Python :: 3.5',
    ],

    ext_modules = [
        Extension('fdpass', ['fdpass.c',]),
    ],

    # What does your project relate to?
    keywords='Mailing list posting confirmation daemon',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #py_modules=["debug"],

    # Setup requirements
    setup_requires=['config', ],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['twine',],
        #'test': ['coverage'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    #package_data={
    #},

    # If set to True, this tells setuptools to automatically include any data
    # files it finds inside your package directories that are specified by your
    # MANIFEST.in file. For more information, see the section below on Including
    # Data Files.
    include_package_data=True,
    
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    data_files=[
        ('/usr/share/man/man1', ['postconfirmc.1',]),
        ('/usr/share/man/man8', ['postconfirmd.8',]),
    ]+conf_files,

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'postconfirmd=postconfirm.postconfirmd:run',
        ],
    },

    scripts = [
        'postconfirmc',
    ],

    # We're reading schema files from a package directory.
    zip_safe = True,

    cmdclass = {
        'install': PostInstallCommand,
    },
)
